"""
Different functions for plotting solver results with Matplotlib.

For the plotter, the following plots are available:
    - plot of the convergence of the matrix solver
    - histogram of the final residuum of the solution
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import matplotlib.pyplot as plt


def _get_plot_residuum(fig, res, data_plot):
    """
    Plot the final residuum (absolute value) with a histogram.
    """

    # extract the data
    n_bins = data_plot["n_bins"]
    tol_bins = data_plot["tol_bins"]
    bar_color = data_plot["bar_color"]
    edge_color = data_plot["edge_color"]

    # activate the figure
    plt.figure(fig)

    # get rms
    n_dof = len(res)
    res_rms = np.sqrt(np.mean(np.abs(res)**2))

    # get absolute value
    res_abs = np.abs(res)

    # clamp small and large values
    v_min = np.finfo(res_abs.dtype).eps
    v_max = np.finfo(res_abs.dtype).max
    res_abs = np.clip(res_abs, v_min, v_max)

    # get the bins
    v_min = np.min(res_abs)/(1+tol_bins)
    v_max = np.max(res_abs)*(1+tol_bins)
    bins = np.logspace(np.log10(v_min), np.log10(v_max), n_bins)

    # plot the histogram
    plt.hist(res_abs, bins=bins, edgecolor=edge_color, color=bar_color)

    # get log axis
    plt.xscale("log")
    plt.yscale("log")

    # add cosmetics
    plt.grid()
    plt.xlabel("residuum (a.u.)")
    plt.ylabel("counts (a.u.)")
    plt.title("Residuum / n_dof = %d / res_rms = %.2e" % (n_dof, res_rms))


def _get_plot_convergence(fig, conv, data_plot):
    """
    Plot the convergence of the iterative matrix solver.
    """

    # extract the data
    color_active = data_plot["color_active"]
    color_reactive = data_plot["color_reactive"]
    marker = data_plot["marker"]
    width = data_plot["width"]

    # activate the figure
    plt.figure(fig)

    # counts
    power_init = conv["power_init"]
    power_final = conv["power_final"]
    power_vec = conv["power_vec"]

    # get convergence
    power_vec = np.concatenate(([power_init], power_vec))
    error_vec = (power_vec-power_final)/np.abs(power_final)
    real_vec = np.abs(np.real(error_vec))
    imag_vec = np.abs(np.imag(error_vec))

    # clamp small and large values
    v_min = np.finfo(error_vec.dtype).eps
    v_max = np.finfo(error_vec.dtype).max
    real_vec = np.clip(real_vec, v_min, v_max)
    imag_vec = np.clip(imag_vec, v_min, v_max)

    # plot the data
    plt.plot(real_vec, "-o", color=color_active, markersize=marker, linewidth=width, label="P")
    plt.plot(imag_vec, "-o", color=color_reactive, markersize=marker, linewidth=width, label="Q")

    # get log axis
    plt.yscale("log")

    # add cosmetics
    plt.grid()
    plt.legend()
    plt.xlabel("iterations (#)")
    plt.ylabel("convergence (a.u.)")
    plt.title(f"Convergence: S = {power_final:.2e} VA")


def get_plot_plotter(fig, res, conv, layout, data_plot, data_options):
    """
    Plot the solver status (for the plotter).
    """

    # extract the data
    style = data_options["style"]
    legend = data_options["legend"]
    font = data_options["font"]

    # plot parameters
    param = {"font.size": font, "legend.loc": legend}

    # get the main plot
    with plt.style.context(style):
        with plt.rc_context(param):
            if layout == "convergence":
                _get_plot_convergence(fig, conv, data_plot)
            elif layout == "residuum":
                _get_plot_residuum(fig, res, data_plot)
            else:
                raise ValueError("invalid plot type and plot feature")



"""
Different functions for resampling a voxel structure.
Refine and reduce a voxel structure and update the voxel indices.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
from pypeec import log

# get a logger
LOGGER = log.get_logger("SAMPLE")


def _get_idx_resample_tensor(resampling_factor, idx_n, idx_r, idx):
    """
    Transform original voxel indices into resampled voxel indices.

    The input indices are linear indices.
    The output indices are tensor indices.
    """

    # extract the resampling
    (rx, ry, rz) = resampling_factor

    # get the provided indices (tensor indices instead of linear indices)
    idx_n_x = idx_n[:, 0][idx]
    idx_n_y = idx_n[:, 1][idx]
    idx_n_z = idx_n[:, 2][idx]

    # get the tensor indices of a single resampled voxel
    idx_r_x = idx_r[:, [0]]
    idx_r_y = idx_r[:, [1]]
    idx_r_z = idx_r[:, [2]]

    # get the global indices
    idx_x = rx*idx_n_x+idx_r_x
    idx_y = ry*idx_n_y+idx_r_y
    idx_z = rz*idx_n_z+idx_r_z

    return idx_x, idx_y, idx_z


def _get_idx_reduce_tensor(idx_n, idx_min, idx):
    """
    Transform original voxel indices into reduced tensor indices.

    The input indices are linear indices.
    The output indices are tensor indices.
    """

    # get the provided indices (tensor indices instead of linear indices)
    idx_x = idx_n[:, 0][idx]
    idx_y = idx_n[:, 1][idx]
    idx_z = idx_n[:, 2][idx]

    # reduce the size of the voxel structure
    idx_x -= idx_min[0]
    idx_y -= idx_min[1]
    idx_z -= idx_min[2]

    return idx_x, idx_y, idx_z


def _get_idx_linear(n, idx_x, idx_y, idx_z):
    """
    Transform tensor indices into linear indices.
    """

    # extract the voxel data
    (nx, ny, nz) = n

    # convert tensor indices into linear indices
    idx_out = idx_x+nx*idx_y+nx*ny*idx_z
    idx_out = idx_out.flatten()

    return idx_out


def _get_original_grid(n):
    """
    Get the indices of the voxels composing the original grid.
    The first resampled voxel has the index zero.
    """

    # extract the voxel data
    (nx, ny, nz) = n

    # get the total number of voxels
    nv = nx*ny*nz

    # get the indices of the original grid
    idx_all = np.arange(nv, dtype=np.int_)
    (idx_n_x, idx_n_y, idx_n_z) = np.unravel_index(idx_all, (nx, ny, nz), order="F")

    # assemble the coordinate array
    idx_n = np.stack((idx_n_x, idx_n_y, idx_n_z), axis=1)

    return idx_n


def _get_resampled_grid(resampling_factor):
    """
    Get the indices of the different sub-voxels composing a single original voxel (after resampling).
    The first resampled sub-voxel has the index zero.
    """

    # extract the resampling resampling_factors
    (rx, ry, rz) = resampling_factor

    # get the number of resampled voxels per original voxel
    rv = rx*ry*rz

    # get the indices of a single resampled voxel
    idx_all = np.arange(rv, dtype=np.int_)
    (idx_r_x, idx_r_y, idx_r_z) = np.unravel_index(idx_all, (rx, ry, rz), order="F")

    # assemble the coordinate array
    idx_r = np.stack((idx_r_x, idx_r_y, idx_r_z), axis=1)

    return idx_r


def _get_grid_bounds(idx_n, domain_def):
    """
    Get the bounds of the tensor structures.
    The bounds are meant with respect to the non-empty voxels.
    """

    # init the array
    idx_lin = np.empty(0, dtype=np.int_)

    # find the indices
    for tag, idx in domain_def.items():
        idx_lin = np.append(idx_lin, idx)

    # get the tensor indices
    idx_grid = idx_n[idx_lin]

    # get the bounds
    idx_min = np.min(idx_grid, axis=0)
    idx_max = np.max(idx_grid, axis=0)

    return idx_min, idx_max


def _get_reduce_indices(n, idx_n, idx_min, domain_def):
    """
    Update the indices of the domains such that they match the reduced structure.
    """

    for tag, idx in domain_def.items():
        # update the old voxel indices into the new tensor indices
        (idx_x, idx_y, idx_z) = _get_idx_reduce_tensor(idx_n, idx_min, idx)

        # transform the tensor indices into linear indices
        idx = _get_idx_linear(n, idx_x, idx_y, idx_z)

        # assign the new indices
        domain_def[tag] = idx

    return domain_def


def _get_update_indices(n, resampling_factor, idx_n, idx_r, domain_def):
    """
    Update the indices of the domains such that they match the resampled structure.
    """

    for tag, idx in domain_def.items():
        # update the old voxel indices into the new tensor indices
        (idx_x, idx_y, idx_z) = _get_idx_resample_tensor(resampling_factor, idx_n, idx_r, idx)

        # transform the tensor indices into linear indices
        idx = _get_idx_linear(n, idx_x, idx_y, idx_z)

        # assign the new indices
        domain_def[tag] = idx

    return domain_def


def _get_update_size(n, d, resampling_factor):
    """
    Update the number of voxels and the dimension of the voxels for the given resampling.
    """

    # extract the voxel data
    (nx, ny, nz) = n
    (dx, dy, dz) = d

    # extract the resampling resampling_factors
    (rx, ry, rz) = resampling_factor

    # update the number of voxels
    n = [nx*rx, ny*ry, nz*rz]

    # update the dimension of the voxels
    d = [dx/rx, dy/ry, dz/rz]

    return n, d


def _get_reduce(n, d, c, domain_def):
    """
    Remove unused voxels from a voxel structure.
    """

    # get the original grid indices
    idx_n = _get_original_grid(n)

    # update the indices of the problem
    (idx_min, idx_max) = _get_grid_bounds(idx_n, domain_def)

    # get the center shift
    n_diff = (idx_min+idx_max+1-n)/2
    c += n_diff*d

    # get the new voxel number
    n = idx_max-idx_min+1

    # cast to list
    n = n.tolist()
    c = c.tolist()

    # update the indices of the problem
    domain_def = _get_reduce_indices(n, idx_n, idx_min, domain_def)

    return n, c, domain_def


def _get_resample(n, d, domain_def, resampling_factor):
    """
    Resampling of a voxel structure (increases the number of voxels).
    """

    # get the original grid indices
    idx_n = _get_original_grid(n)

    # get the indices of a single resampled voxel
    idx_r = _get_resampled_grid(resampling_factor)

    # update the voxel number and size
    (n, d) = _get_update_size(n, d, resampling_factor)

    # update the indices of the problem
    domain_def = _get_update_indices(n, resampling_factor, idx_n, idx_r, domain_def)

    return n, d, domain_def


def get_remesh(n, d, c, domain_def, resampling):
    """
    Remesh of a voxel structure (remove unused voxels and resampling).
    """

    # extract the data
    use_reduce = resampling["use_reduce"]
    use_resample = resampling["use_resample"]
    resampling_factor = resampling["resampling_factor"]

    # display number of voxels
    LOGGER.debug("use_reduce = %s" % use_reduce)
    LOGGER.debug("use_resample = %s" % use_resample)
    LOGGER.debug("original number = %d" % np.prod(n))

    # remove unused voxels
    if use_reduce:
        (n, c, domain_def) = _get_reduce(n, d, c, domain_def)

    # resampling of the voxel structure
    if use_resample:
        (n, d, domain_def) = _get_resample(n, d, domain_def, resampling_factor)

    # voxel structure size
    s = tuple(x*y for x, y in zip(n, d))

    # display number of voxels
    LOGGER.debug("final number = %d" % np.prod(n))

    return n, d, c, s, domain_def



"""
Different functions for building the equation system.
Two equation systems are built, one for the preconditioner and one for the full system.

The voxel structure has the following size: (nx, ny, nz).
The problem contains n_vc non-empty electric voxels and n_vm non-empty magnetic voxels.
The problem contains n_fc internal electric faces and n_fm internal magnetic faces.
The problem contains n_src_c current source voxels and n_src_v voltage source voxels.

The equations are set in the following order:
    - n_fc: the electric KVL equations
    - n_vc: the electric KCL equations
    - n_src_c: the current source voxels equations (source equation with internal admittance)
    - n_src_v: the voltage source voxels equations (source equation with internal impedance)
    - n_fm: the magnetic KVL equations
    - n_vm: the magnetic KCL equations

The complete equation matrix is:
    [
        +R_c+s*L_c,    -A_net_c',     0,            +K_c,             0 ;
        +A_net_c,       0,           +A_vc_src,      0,               0 ;
         0,            +A_src_vc,    +A_src_src,     0,               0 ;
        -K_m,           0,            0,            +R_m,            -A_net_m' ;
         0,             0,            0,            +P_m*A_net_m,    +s*I ;
    ]

The complete solution vector is:
    [
        n_fc: electric face currents
        n_vc: electric voxel potentials
        n_src_c: current source currents
        n_src_v: voltage source currents
        n_fm: magnetic face fluxes
        n_vm: magnetic voxel potentials
    ]

The complete right-hand size vector is:
    [
        n_fc: zero excitation
        n_vc: zero excitation
        n_src_c: current source excitations
        n_src_v: voltage source excitations
        n_fm: zero excitation
        n_vm: zero excitation
    ]

For the DC problem (zero frequency), division per zero are occurring.
Therefore, the problem is formulated slightly differently for this case.

For the preconditioner, the diagonal of the inductance and potential matrix is used.
For the preconditioner, the electric-magnetic coupling matrices are simplified.
The preconditioner is solved with the Schur complement and the matrix factorization.
The preconditioner is solved separately for the electric and magnetic equations.

The preconditioner matrices (electric and magnetic) have the following form:
    [
        Z_mat,       A_12_mat;
        A_21_mat,    A_22_mat;
    ]

The matrix impedance matrix (Z_mat) is diagonal.
Therefore, the factorization is computed on the Schur complement.
The Schur complement is computed as:
        - Y_mat = 1/Z_mat
        - S_mat = A_22_mat-A_21_mat*Y_mat*A_12_mat

Two different methods are available to factorize and solve the Schur complement.

For the full equation system, the complete dense matrix are used:
    - the system is split in three parts: electric, magnetic, and electric-magnetic coupling
    - the system is meant to be solved with an iterative solver
    - the full system matrix is not built and a matrix-vector operator is returned
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import scipy.sparse as sps
from pypeec.lib_matrix import matrix_factorization


def _get_system_size(A_net_c, A_net_m, A_src):
    """
    Get the size of the equation system.

    The equation system has the following size:
        - n_fc+n_vc+n_src_c+n_src_v: electric system
        - n_fm+n_vm: magnetic system
    """

    # get the matrices
    (A_vc_src, A_src_vc, A_src_src) = A_src

    # get the system size
    (n_vc, n_fc) = A_net_c.shape
    (n_vm, n_fm) = A_net_m.shape
    (n_src, n_src) = A_src_src.shape

    return n_vc, n_fc, n_vm, n_fm, n_src


def _get_system_scaler(freq, n_vc, n_fc, n_vm, n_fm, n_src):
    """
    Get a scaling vector for the solution.

    The scaling vector is required for the following reasons:
        - for avoid division per zero, the DC and AC system are slightly different
        - the DC system is using the magnetic flux as a variable
        - the AC system is using the derivative of the magnetic flux as a variable
        - the solution is using the magnetic flux as a variable
        - the scaling vector is used to match this discrepancy
    """

    # get the angular frequency
    s = 1j*2*np.pi*freq

    # get the scaling vectors
    if freq == 0:
        scaler_c = np.ones(n_fc+n_vc+n_src, dtype=np.complex_)
        scaler_m = np.ones(n_fm+n_vm, dtype=np.complex_)
    else:
        scaler_c = np.ones(n_fc+n_vc+n_src, dtype=np.complex_)
        scaler_fm = s*np.ones(n_fm, dtype=np.complex_)
        scaler_vm = np.ones(n_vm, dtype=np.complex_)
        scaler_m = np.concatenate((scaler_fm, scaler_vm))

    return scaler_c, scaler_m


def _get_coupling_electric(sol_m, freq, n_vc, n_fc, n_fm, n_src, K_op_c):
    """
    Compute the magnetic to electric couplings.
    The magnetic face currents are multiplied with the couplings matrices.
    The coupling contributions to the electric rhs vector are returned.

    The vectors have the following size: n_fc+n_vc+n_src_c+n_src_v.
    """

    # extract the face current
    I_fm = sol_m[0:n_fm]

    # compute the couplings
    if freq == 0:
        cpl_fc = np.zeros(n_fc, dtype=np.complex_)
    else:
        cpl_fc = K_op_c(I_fm)

    cpl_vc = np.zeros(n_vc, dtype=np.complex_)
    cpl_src = np.zeros(n_src, dtype=np.complex_)

    # assemble the vectors
    cpl_c = np.concatenate((cpl_fc, cpl_vc, cpl_src))

    return cpl_c


def _get_coupling_magnetic(sol_c, n_fc, n_vm, K_op_m):
    """
    Compute the electric to magnetic couplings.
    The electric face currents are multiplied with the couplings matrices.
    The coupling contributions to the magnetic rhs vector are returned.

    The vectors have the following size: n_fm+n_vm.
    """

    # extract the face current
    I_fc = sol_c[0:n_fc]

    # compute the couplings
    cpl_fm = -K_op_m(I_fc)
    cpl_vm = np.zeros(n_vm, dtype=np.complex_)

    # assemble the vectors
    cpl_m = np.concatenate((cpl_fm, cpl_vm))

    return cpl_m


def _get_cond_fact_electric(freq, A_net_c, R_c, L_c, A_src):
    """
    Compute the sparse matrices using for the electric preconditioner.

    The equation system has the following size: n_fc+n_vc+n_src_c+n_src_v.
    The Schur complement split the system in two subsystems:
        - n_fc (diagonal block size)
        - n_vc+n_src_c+n_src_v (Schur complement size)
    """

    # get the matrices
    (A_vc_src, A_src_vc, A_src_src) = A_src

    # get the system size
    (n_vc, n_fc) = A_net_c.shape
    (n_src, n_src) = A_src_src.shape

    # get the angular frequency
    s = 1j*2*np.pi*freq

    # get the electric admittance
    Y_c = 1/(R_c+s*L_c)

    # admittance matrix
    Y_mat = sps.diags(Y_c, format="csc")

    # assemble the matrices
    A_12_mat = -A_net_c.transpose()
    A_21_mat = A_net_c
    A_22_mat = sps.csc_matrix((n_vc, n_vc), dtype=np.int_)

    # expand for the source matrices
    A_add = sps.csc_matrix((n_fc, n_src), dtype=np.int_)
    A_12_mat = sps.hstack([A_12_mat, A_add], dtype=np.complex_, format="csr")

    # expand for the source matrices
    A_add = sps.csc_matrix((n_src, n_fc), dtype=np.int_)
    A_21_mat = sps.vstack([A_21_mat, A_add], dtype=np.complex_, format="csc")

    # add the source
    A_22_mat = sps.bmat([[A_22_mat, A_vc_src], [A_src_vc, A_src_src]], dtype=np.complex_, format="csc")

    # computing the Schur complement (with respect to the diagonal admittance matrix)
    S_mat = A_22_mat-A_21_mat*Y_mat*A_12_mat

    return Y_mat, S_mat, A_12_mat, A_21_mat


def _get_cond_fact_magnetic(freq, A_net_m, R_m, P_m):
    """
    Compute the sparse matrices using for the magnetic preconditioner.

    The equation system has the following size: n_fm+n_vm.
    The Schur complement split the system in two subsystems:
        - n_fm (diagonal block size)
        - n_vm (Schur complement size)
    """

    # get the system size
    (n_vm, n_fm) = A_net_m.shape

    # get the angular frequency
    s = 1j*2*np.pi*freq

    # get the magnetic admittance (avoid singularity for DC solution)
    if freq == 0:
        Y_m = 1/R_m
        I_m = np.ones(n_vm, dtype=np.complex_)
    else:
        Y_m = s/R_m
        I_m = s*np.ones(n_vm, dtype=np.complex_)

    # admittance matrix
    Y_mat = sps.diags(Y_m, format="csc")

    # potential matrix
    I_mat_m = sps.diags(I_m, format="csc")

    # assemble the matrices
    A_12_mat = -A_net_m.transpose()
    A_21_mat = P_m*A_net_m
    A_22_mat = I_mat_m

    # computing the Schur complement (with respect to the diagonal admittance matrix)
    S_mat = A_22_mat-A_21_mat*Y_mat*A_12_mat

    return Y_mat, S_mat, A_12_mat, A_21_mat


def _get_cond_solve(rhs, Y_mat, S_fact, A_12_mat, A_21_mat):
    """
    Solve the preconditioner equation system.
    The matrix factorization of the Schur complement is used.
    """

    # split the system (Schur complement split)
    (n_schur, n_schur) = Y_mat.shape

    # split the rhs vector
    rhs_a = rhs[:n_schur]
    rhs_b = rhs[n_schur:]

    # solve the equation system (Schur complement and matrix factorization)
    tmp = rhs_b-(A_21_mat*(Y_mat*rhs_a))
    sol_b = matrix_factorization.get_solve(S_fact, tmp)
    sol_a = Y_mat*(rhs_a-(A_12_mat*sol_b))

    # assemble the solution
    sol = np.concatenate((sol_a, sol_b))

    return sol


def _get_system_multiply_electric(sol, freq, A_net_c, A_src, R_c, L_op_c):
    """
    Multiply the full electric equation matrix with a given solution test vector.

    For the multiplication of the dense matrix, the provided linear operators are used.
    The equation system has the following size: n_fc+n_vc+n_src_c+n_src_v.
    """

    # get the matrices
    (A_vc_src, A_src_vc, A_src_src) = A_src

    # get the system size
    (n_vc, n_fc) = A_net_c.shape
    (n_src, n_src) = A_src_src.shape

    # get the angular frequency
    s = 1j*2*np.pi*freq

    # split the solution vector
    I_fc = sol[0:n_fc]
    V_vc = sol[n_fc:n_fc+n_vc]
    I_src = sol[n_fc+n_vc:n_fc+n_vc+n_src]

    # multiply the inductance matrix
    if freq == 0:
        rhs_kvl_ind = np.zeros(n_fc, dtype=np.complex_)
    else:
        rhs_kvl_ind = s*L_op_c(I_fc)

    # electric KVL equations
    rhs_kvl_res = R_c*I_fc
    rhs_kvl_net = -A_net_c.transpose()*V_vc

    # electric KCL equations
    rhs_kcl_net = A_net_c*I_fc
    rhs_kvl_src = A_vc_src*I_src

    # form the source equation
    rhs_src_con = A_src_vc*V_vc
    rhs_src_src = A_src_src*I_src

    # assemble the solution
    rhs_kvl = rhs_kvl_ind+rhs_kvl_res+rhs_kvl_net
    rhs_kcl = rhs_kcl_net+rhs_kvl_src
    rhs_src = rhs_src_con+rhs_src_src
    rhs = np.concatenate((rhs_kvl, rhs_kcl, rhs_src))

    return rhs


def _get_system_multiply_magnetic(sol, freq, A_net_m, R_m, P_op_m):
    """
    Multiply the full magnetic equation matrix with a given solution test vector.

    For the multiplication of the dense matrix, the provided linear operators are used.
    The equation system has the following size: n_fm+n_vm.
    """

    # get the system size
    (n_vm, n_fm) = A_net_m.shape

    # get the angular frequency
    s = 1j*2*np.pi*freq

    # split the solution vector
    I_fm = sol[0:n_fm]
    V_vm = sol[n_fm:n_fm+n_vm]

    # multiply the potential matrix
    rhs_kcl_pot = P_op_m(A_net_m*I_fm)

    # get the term that are different for DC and AC cases
    if freq == 0:
        rhs_kvl_res = R_m*I_fm
        rhs_kcl_net = V_vm
    else:
        rhs_kvl_res = R_m/s*I_fm
        rhs_kcl_net = s*V_vm

    # magnetic KVL equations
    rhs_kvl_net = -A_net_m.transpose()*V_vm

    # assemble the solution
    rhs_kvl = rhs_kvl_res+rhs_kvl_net
    rhs_kcl = rhs_kcl_pot+rhs_kcl_net
    rhs = np.concatenate((rhs_kvl, rhs_kcl))

    return rhs


def get_source_vector(idx_vc, idx_vm, idx_fc, idx_fm, I_src_c, V_src_v):
    """
    Construct the right-hand side with the current and voltage sources.

    The right-hand size vector has the following size: n_fc+n_fm+n_vc+n_vm+n_src_c+n_src_v.
    The excitations are only located in the last equations: n_src_c+n_src_v.
    """

    # extract the voxel data
    n_c = len(idx_vc)+len(idx_fc)
    n_m = len(idx_vm)+len(idx_fm)

    # excitation are handled separately
    rhs_c = np.zeros(n_c, dtype=np.complex_)
    rhs_m = np.zeros(n_m, dtype=np.complex_)

    # assemble
    rhs_c = np.concatenate((rhs_c, I_src_c, V_src_v))

    return rhs_c, rhs_m


def get_source_matrix(idx_vc, idx_src_c, idx_src_v, G_src_c, R_src_v):
    """
    Construct the source matrices.
    The source matrices describes the sources (internal resistances/admittances).
    The source matrices connect the sources to the rest of the system.

    The source matrices have the following sizes:
        - A_vc_src: (n_vc, n_src_c+n_src_v)
        - A_src_vc: (n_src_c+n_src_v, n_vc)
        - A_src_src: (n_src_c+n_src_v, n_src_c+n_src_v)

    It should be noted that the matrices connecting the magnetic voxels are all-zeros.
    This is explained by the fact that the sources are only connected to electric voxels.
    """

    # extract the voxel data
    n_vc = len(idx_vc)
    n_src_c = len(idx_src_c)
    n_src_v = len(idx_src_v)

    # find the variable indices
    idx_s = np.argsort(idx_vc)
    idx_src_c_p = np.searchsorted(idx_vc, idx_src_c, sorter=idx_s)
    idx_src_v_p = np.searchsorted(idx_vc, idx_src_v, sorter=idx_s)
    idx_src_c_local = idx_s[idx_src_c_p]
    idx_src_v_local = idx_s[idx_src_v_p]

    # indices of the new source equations to be added
    idx_src_c_add = np.arange(0, n_src_c, dtype=np.int_)
    idx_src_v_add = np.arange(n_src_c, n_src_c+n_src_v, dtype=np.int_)

    # constant vector with the size of the sources
    cst_src_c = np.full(n_src_c, 1, dtype=np.complex_)
    cst_src_v = np.full(n_src_v, 1, dtype=np.complex_)

    # matrix between the KCL equations and the source variables
    idx_row = np.concatenate((idx_src_c_local, idx_src_v_local))
    idx_col = np.concatenate((idx_src_c_add, idx_src_v_add))
    val = np.concatenate((-cst_src_c, -cst_src_v))
    A_vc_src = sps.csc_matrix((val, (idx_row, idx_col)), shape=(n_vc, n_src_c+n_src_v), dtype=np.complex_)

    # matrix between the source equations and the potential variables
    idx_row = np.concatenate((idx_src_v_add, idx_src_c_add))
    idx_col = np.concatenate((idx_src_v_local, idx_src_c_local))
    val = np.concatenate((cst_src_v, G_src_c))
    A_src_vc = sps.csc_matrix((val, (idx_row, idx_col)), shape=(n_src_c+n_src_v, n_vc), dtype=np.complex_)

    # matrix between the source equations and the source variables
    idx_row = np.concatenate((idx_src_c_add, idx_src_v_add))
    idx_col = np.concatenate((idx_src_c_add, idx_src_v_add))
    val = np.concatenate((cst_src_c, R_src_v))
    A_src_src = sps.csc_matrix((val, (idx_row, idx_col)), shape=(n_src_c+n_src_v, n_src_c+n_src_v), dtype=np.complex_)

    return A_vc_src, A_src_vc, A_src_src


def get_cond_operator(freq, A_net_c, A_net_m, A_src, R_c, R_m, L_c, P_m):
    """
    Get linear operators that solves the preconditioner equation system.
    These operators are used as a preconditioner for the iterative method solving the full system.

    The system is split between the electric and magnetic equations.
    The Schur complement matrices of the preconditioner are also returned.
    These matrices are used to assess the condition number of the system.
    """

    # get the system size and the solution scaling
    (n_vc, n_fc, n_vm, n_fm, n_src) = _get_system_size(A_net_c, A_net_m, A_src)
    (scaler_c, scaler_m) = _get_system_scaler(freq, n_vc, n_fc, n_vm, n_fm, n_src)

    # get the Schur complement
    (Y_mat_c, S_mat_c, A_12_mat_c, A_21_mat_c) = _get_cond_fact_electric(freq, A_net_c, R_c, L_c, A_src)
    (Y_mat_m, S_mat_m, A_12_mat_m, A_21_mat_m) = _get_cond_fact_magnetic(freq, A_net_m, R_m, P_m)

    # factorize the Schur
    S_fact_c = matrix_factorization.get_factorize("electric", S_mat_c)
    S_fact_m = matrix_factorization.get_factorize("magnetic", S_mat_m)

    # function describing the electric preconditioner
    def fct_c(rhs_c):
        # solve the system (electric and magnetic)
        sol_c = _get_cond_solve(rhs_c, Y_mat_c, S_fact_c, A_12_mat_c, A_21_mat_c)

        # scale and assemble the solution
        sol_c = sol_c/scaler_c

        return sol_c

    # function describing the magnetic preconditioner
    def fct_m(rhs_m):
        # solve the system (electric and magnetic)
        sol_m = _get_cond_solve(rhs_m, Y_mat_m, S_fact_m, A_12_mat_m, A_21_mat_m)

        # scale and assemble the solution
        sol_m = sol_m/scaler_m

        return sol_m

    # corresponding linear operator
    fct = (fct_c, fct_m)

    return fct, S_mat_c, S_mat_m


def get_coupling_operator(freq, A_net_c, A_net_m, A_src, K_op_c, K_op_m):
    """
    Get linear operators that represent the electric-magnetic couplings.

    The system is split between the electric and magnetic equations.
    These operators are coupling both systems.
    """

    # get the system size and the solution scaling
    (n_vc, n_fc, n_vm, n_fm, n_src) = _get_system_size(A_net_c, A_net_m, A_src)
    (scaler_c, scaler_m) = _get_system_scaler(freq, n_vc, n_fc, n_vm, n_fm, n_src)

    # function describing the electric couling
    def fct_c(sol_m):
        # scale equation
        sol_m = sol_m*scaler_m

        # compute the electric-magnetic coupling
        cpl_c = _get_coupling_electric(sol_m, freq, n_vc, n_fc, n_fm, n_src, K_op_c)

        return cpl_c

    # function describing the magnetic preconditioner
    def fct_m(sol_c):
        # scale equation
        sol_c = sol_c*scaler_c

        # compute the electric-magnetic coupling
        cpl_m = _get_coupling_magnetic(sol_c, n_fc, n_vm, K_op_m)

        return cpl_m

    # corresponding linear operator
    fct = (fct_c, fct_m)

    return fct


def get_system_operator(freq, A_net_c, A_net_m, A_src, R_c, R_m, L_op_c, P_op_m):
    """
    Get linear operators that produce the matrix-vector multiplication result for the full system.
    These operators are used for the iterative solver.

    The system is split between the electric and magnetic equations.
    """

    # get the system size and the solution scaling
    (n_vc, n_fc, n_vm, n_fm, n_src) = _get_system_size(A_net_c, A_net_m, A_src)
    (scaler_c, scaler_m) = _get_system_scaler(freq, n_vc, n_fc, n_vm, n_fm, n_src)

    # function describing the electric equation system
    def fct_c(sol_c):
        # scale the solution
        sol_c = sol_c*scaler_c

        # compute the system multiplication
        rhs_c = _get_system_multiply_electric(sol_c, freq, A_net_c, A_src, R_c, L_op_c)

        return rhs_c

    # function describing the magnetic equation system
    def fct_m(sol_m):
        # scale the solution
        sol_m = sol_m*scaler_m

        # compute the system multiplication
        rhs_m = _get_system_multiply_magnetic(sol_m, freq, A_net_m, R_m, P_op_m)

        return rhs_m

    # corresponding linear operator
    fct = (
        fct_c,
        fct_m,
    )

    return fct


def get_system_sol_idx(A_net_c, A_net_m, A_src):
    """
    Get the indices of the vectors composing the solution.
    """

    # get the system size and the solution scaling
    (n_vc, n_fc, n_vm, n_fm, n_src) = _get_system_size(A_net_c, A_net_m, A_src)

    # init index dict
    sol_idx = {}

    # init the index offset
    n_offset = 0

    # get the electric variable
    sol_idx["I_fc"] = range(n_offset, n_offset+n_fc)
    n_offset += n_fc
    sol_idx["V_vc"] = range(n_offset, n_offset+n_vc)
    n_offset += n_vc
    sol_idx["I_src"] = range(n_offset, n_offset+n_src)
    n_offset += n_src

    # get the magnetic variable
    sol_idx["I_fm"] = range(n_offset, n_offset+n_fm)
    n_offset += n_fm
    sol_idx["V_vm"] = range(n_offset, n_offset+n_vm)
    n_offset += n_vm

    return sol_idx



"""
Generate PNG files from the GERBER files for the example.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import sys
from pypeec.utils import gerber_png


if __name__ == "__main__":
    # ######################## definition of the colors
    color_def = {
        "none":  (255, 255, 255),
        "copper": (255, 0, 0),
        "sink":  (0, 0, 255),
        "src": (0, 255, 0),
    }

    # ######################## definition of the GERBER files
    gerber_def = {
        "front": "pcb-front_copper.gbr",
        "back": "pcb-back_copper.gbr",
        "src": "pcb-src.gbr",
        "sink": "pcb-sink.gbr",
        "drill": "pcb-PTH.drl",
        "via":  "pcb-VIA.drl",
    }

    # ######################## assemble the GERBER data
    data_gerber = {
        "gerber_edge": "pcb-edge.gbr",           # name of the GERBER file with the board edgess
        "color_edge": (0, 0, 0),                 # color of the board edges
        "color_background": (255, 255, 255),     # background color
        "color_def": color_def,                  # dict with the color definition
        "gerber_def": gerber_def                 # dict with the GERBER file definition
    }

    # ######################## get the layer stack (combination between GERBER files and colors)
    data_stack = {
        "front": [
            {"gerber": "drill", "color": "none"},
            {"gerber": "front", "color": "copper"},
        ],
        "back": [
            {"gerber": "drill", "color": "none"},
            {"gerber": "back", "color": "copper"},
        ],
        "terminal": [
            {"gerber": "drill", "color": "none"},
            {"gerber": "sink", "color": "sink"},
            {"gerber": "src", "color": "src"},
        ],
        "via": [
            {"gerber": "drill", "color": "none"},
            {"gerber": "via", "color": "copper"},
        ],
    }

    # ######################## get the export instructions
    data_export = {
        "margin": 0.1,                    # relative margin to be considered around the board
        "oversampling": 1.0,              # oversampling factor for exporting the GERBER files
        "voxel": 17.0e-6,                 # size of the voxel
        "folder_gerber": "../gerber",     # GERBER file location
        "folder_png": "../png",           # PNG file location
    }

    # ######################## run
    gerber_png.get_convert(data_export, data_gerber, data_stack)

    # ######################## exit
    sys.exit(0)



"""
Configuration file for the Sphinx documentation generator.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os
import sys
import datetime
import setuptools_scm

# define the package path
sys.path.insert(0, os.path.abspath('..'))

# get the version number
ver = setuptools_scm.get_version(
    root='..',
    relative_to=__file__,
    version_scheme="guess-next-dev",
)
release = ver
version = ver

# get date
date = datetime.datetime.today().strftime('%a, %b %d, %Y')

# project metadata
project = 'PyPEEC'
html_title = "PyPEEC"
author = 'Thomas Guillod'
copyright = 'Thomas Guillod - Dartmouth College'

# add prolog
rst_epilog = """
.. |ver| replace:: {ver}
.. |date| replace:: {date}
"""
rst_epilog = rst_epilog.format(ver=ver, date=date)

# load extensions
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
]

# autodoc order
autodoc_member_order = 'bysource'

# disable API file generation
autosummary_generate = False

# allowed redirect for the external links
linkcheck_allowed_redirects = {
    'http://mozilla.org/MPL/2.0': 'https://www.mozilla.org/en-US/MPL/2.0',
    'https://dartmouth.edu': 'https://home.dartmouth.edu'
}

# define paths
templates_path = ['_templates']
html_static_path = ['_static']
html_theme_path = ['_static']

# html options
html_baseurl = 'https://pypeec.otvam.ch'
html_theme = 'sphinx_rtd_theme'
html_logo = "images/sphinx.png"
html_favicon = "images/icon.png"
html_show_sphinx = False
html_show_sourcelink = False
html_copy_source = False
html_use_index = False
html_domain_indices = False

# html theme
html_theme_options = {
    'logo_only': True,
    'display_version': True,
    'style_nav_header_background': '#137e9c',
    'prev_next_buttons_location': None,
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 2,
    'includehidden': True,
    'titles_only': False,
}



"""
Different functions for checking the point cloud.
The point cloud should not intersect with the non-empty voxels.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
from pypeec import log

# get a logger
LOGGER = log.get_logger("CLOUD")


def _get_voxel_coordinate(n, d, c, idx_all):
    """
    Get the coordinate of the different voxels.
    The center voxel center is at the specified origin coordinate.
    """

    # convert linear indices into tensor indices
    (idx_x, idx_y, idx_z) = np.unravel_index(idx_all, n, order="F")

    # assemble the coordinate array
    idx_vox = np.stack((idx_x, idx_y, idx_z), axis=1)

    # origin coordinate
    o = c-(n*d)/2

    # assemble the coordinate array
    pts_vox = o+d/2+d*idx_vox

    return pts_vox


def _get_point_valid(d, pts_vox, pts_tmp):
    """
    Check of a cloud point is not intersecting with the non-empty voxels.
    """

    # compute distance for each dimension
    pts_dis = np.abs(pts_vox-pts_tmp)
    pts_valid = pts_dis > d

    # check if distance are respected
    pts_valid = np.any(pts_valid == True, axis=1)
    valid = np.all(pts_valid == True, axis=0)

    return valid


def get_cloud(n, d, c, domain_def, pts_cloud_in):
    """
    Check of the cloud points are not intersecting with the non-empty voxels.
    Keep the valid points, discard the other points.
    """

    # display number of cloud points
    LOGGER.debug("initial number = %d" % len(pts_cloud_in))

    # cast to array
    c = np.array(c, dtype=np.float_)
    d = np.array(d, dtype=np.float_)
    n = np.array(n, dtype=np.int_)

    # assemble all the indices
    idx_all = np.empty(0, dtype=np.int_)
    for idx in domain_def.values():
        idx_all = np.append(idx_all, idx)

    # get the voxel coordinate
    pts_vox = _get_voxel_coordinate(n, d, c, idx_all)

    # check the points
    pts_cloud_out = []
    for pts_tmp in pts_cloud_in:
        valid = _get_point_valid(d, pts_vox, pts_tmp)
        if valid:
            pts_cloud_out.append(pts_tmp)

    # cast
    pts_cloud_out = np.array(pts_cloud_out, np.float_)

    # display number of cloud points
    LOGGER.debug("final number = %d" % len(pts_cloud_out))

    return pts_cloud_out



"""
Script for rendering the PyPEEC logo.

Three files are created:
    - logo.ply: 3D logo in PLY format
    - logo.png: 2D logo in PNG format
    - icon.png: 2D icon in PNG format
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import pyvista as pv
import matplotlib.cm as cm
import PIL.Image as pmg


def get_voxel():
    """
    Create the voxel structure with the logo.
    """

    # voxel indices
    idx = [
        0,  1,  2,
        6,  7,  8,
        9, 11, 15,
        18, 20, 23,
        24, 25, 26,
    ]

    # voxel colors
    var = [
        0.85714286, 0.78571429, 0.71428571,
        0.14285714, 0.07142857, 0.00000000,
        0.92857143, 0.64285714, 0.21428571,
        1.00000000, 0.57142857, 0.50000000,
        0.28571429, 0.35714286, 0.42857143,
    ]

    # create a uniform grid for the complete structure
    grid = pv.ImageData()

    # set the array size and the voxel size
    grid.origin = (1.5, 1.5, 1.5)
    grid.dimensions = (4, 4, 4)
    grid.spacing = (1, 1, 1)

    # get the voxel structure
    voxel = grid.extract_cells(idx)
    voxel["var"] = var

    return voxel


def get_plotter(voxel):
    """
    Plotter for rendering the logo from the voxel structure.
    """

    # setup plot
    pv.rcParams['transparent_background'] = True
    pv.rcParams["edge_color"] = "black"

    # create plotter
    pl = pv.Plotter(
        window_size=(256, 256),
        off_screen=True,
        line_smoothing=True,
        polygon_smoothing=True,
        lighting="three lights"
    )

    # add content
    pl.add_mesh(
        voxel,
        scalars="var",
        show_scalar_bar=False,
        show_edges=True,
        edge_color="black",
        cmap="viridis",
        line_width=1.0,
    )

    # set camera
    pl.camera.SetWindowCenter(0, -0.1)
    pl.camera.roll = 0
    pl.camera.azimuth = 180
    pl.camera.elevation = 170

    return pl


def get_mesh(voxel):
    """
    Extract a surface mesh from the voxel structure.
    """

    # get mesh
    mesh = voxel.extract_surface()
    var = mesh["var"]

    # get the texture
    texture = cm.viridis(var)
    texture = 255*texture
    texture = texture.astype(np.uint8)

    return mesh, texture


if __name__ == "__main__":
    # get the data
    voxel = get_voxel()
    pl = get_plotter(voxel)
    (mesh, texture) = get_mesh(voxel)

    # set the filename
    filename_logo_ply = "logo.ply"
    filename_logo_png = "logo.png"
    filename_logo_icn = "icon.png"

    # save the PNG file
    pl.show(screenshot=filename_logo_png)

    # save the PLY file
    mesh.save(filename_logo_ply, texture=texture)

    # create the icon
    img = pmg.open(filename_logo_png, formats=["png"])
    img = img.resize((96, 96), pmg.BICUBIC)
    img.save(filename_logo_icn)



"""
Main script for visualizing a 3D voxel structure.
Plot the following features:
    - the different domain composing the voxel structure
    - the connected components of the voxel structure
    - the deviation between the original geometry and the voxel structure

Three different mode are available for the plots:
    - windowed mode: with the Qt Framework
    - notebook mode: with Jupyter notebook
    - silent mode: nothing is shown and the program exit

The plots are generated with PyVista.
The plots can be saved as PNG images.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_visualization import data_viewer
from pypeec.lib_visualization import data_voxel
from pypeec.lib_visualization import manage_pyvista
from pypeec.lib_visualization import manage_plotgui
from pypeec.lib_check import check_data_visualization
from pypeec.lib_check import check_data_options
from pypeec import log

# get a logger
LOGGER = log.get_logger("VIEWER")


def _get_grid_voxel(data_geom):
    """
    Convert the complete voxel geometry into a PyVista uniform grid.
    Convert the non-empty voxel geometry into a PyVista unstructured grid.
    Add the domain tags to the grid.
    """

    # extract the data
    n = data_geom["n"]
    d = data_geom["d"]
    c = data_geom["c"]
    domain_def = data_geom["domain_def"]
    connection_def = data_geom["connection_def"]
    pts_cloud = data_geom["pts_cloud"]
    reference = data_geom["reference"]

    # get voxel indices
    idx = data_viewer.get_voxel(domain_def)

    # convert the voxel geometry into PyVista grids
    grid = data_voxel.get_grid(n, d, c)
    voxel = data_voxel.get_voxel(grid, idx)
    point = data_voxel.get_point(pts_cloud)
    reference = data_voxel.get_reference(reference, voxel)

    # add the domain tag to the geometry
    voxel = data_viewer.set_data(voxel, idx, domain_def, connection_def)

    return grid, voxel, point, reference


def _get_plot(tag, data_viewer, grid, voxel, point, reference, gui_obj):
    """
    Make a plot with the voxel structure and the domains.
    """

    # extract the data
    framework = data_viewer["framework"]
    title = data_viewer["title"]
    layout = data_viewer["layout"]
    data_window = data_viewer["data_window"]
    data_plot = data_viewer["data_plot"]
    data_options = data_viewer["data_options"]

    # check framework
    if framework != "pyvista":
        raise ValueError("invalid plot framework")

    # get the plotter (with the Qt framework)
    pl = gui_obj.open_pyvista(tag, title, data_window)

    # make the plot
    manage_pyvista.get_plot_viewer(pl, grid, voxel, point, reference, layout, data_plot, data_options)


def run(
        data_voxel, data_viewer,
        tag_plot=None, plot_mode="qt", folder=".", name=None,
):
    """
    Main script for visualizing a 3D voxel structure.
    Handle invalid data with exceptions.
    """

    # check the voxel data
    LOGGER.info("check the voxel data")
    data_geom = check_data_options.check_data_voxel(data_voxel)

    # check the input data
    LOGGER.info("check the input data")
    check_data_visualization.check_data_viewer(data_viewer)
    check_data_options.check_plot_options(plot_mode, folder, name)
    check_data_options.check_tag_list(data_viewer, tag_plot)

    # find the plots
    if tag_plot is not None:
        data_viewer = {key: data_viewer[key] for key in tag_plot}

    # create the Qt app (should be at the beginning)
    LOGGER.info("init the plot manager")
    gui_obj = manage_plotgui.PlotGui(plot_mode, folder, name)

    # handle the data
    LOGGER.info("parse data")
    (grid, voxel, point, reference) = _get_grid_voxel(data_geom)

    # make the plots
    with log.BlockTimer(LOGGER, "generate plots"):
        for i, (tag_plot, data_viewer_tmp) in enumerate(data_viewer.items()):
            LOGGER.info("plotting %d / %d / %s" % (i+1, len(data_viewer), tag_plot))
            _get_plot(tag_plot, data_viewer_tmp, grid, voxel, point, reference, gui_obj)

    # enter the event loop (should be at the end, blocking call)
    gui_obj.show()



"""
Module for transforming STL files into a 3D voxel structure.
Each STL file corresponds to a domain of the 3D voxel structure.

The voxelization is done with PyVista.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import vtk
import numpy as np
import pyvista as pv
from pypeec import log
from pypeec import config

# get a logger
LOGGER = log.get_logger("STL")

# get problem size limits
VOXEL_TOTAL = config.PROBLEM_MAX_SIZE.VOXEL_TOTAL
VOXEL_USED = config.PROBLEM_MAX_SIZE.VOXEL_USED

# prevent VTK to mess up the output
vtk.vtkObject.GlobalWarningDisplayOff()


def _get_load_stl(scale, offset, filename):
    """
    Load several STL files and merge the meshes.
    """

    # load the mesh
    try:
        mesh = pv.read(filename, force_ext=".stl")
    except ValueError:
        raise RuntimeError("invalid stl: invalid file type: %s" % filename)

    # check that the file is valid
    if mesh.n_cells == 0:
        raise RuntimeError("invalid stl: invalid file content: %s" % filename)

    # translate the meshes
    mesh = mesh.scale(scale, inplace=True)
    mesh = mesh.translate(offset, inplace=True)

    return mesh


def _get_voxelize_stl(grid, mesh):
    """
    Voxelize an STL mesh with respect to a uniform grid.
    Return the indices of the created voxels.
    """

    # voxelize the mesh
    try:
        selection = grid.select_enclosed_points(mesh, tolerance=0.0, check_surface=True)
    except RuntimeError:
        raise RuntimeError("invalid mesh: mesh cannot be voxelized")

    # create a boolean mask
    mask = selection["SelectedPoints"].view(bool)

    # find the voxel indices
    if np.any(mask):
        # transform the grid into an unstructured grid (keeping the non-empty voxels)
        voxel = grid.extract_points(mask)

        # get the indices of the extracted voxels
        idx_voxel = voxel["idx"]
    else:
        idx_voxel = np.empty(0, dtype=np.int_)

    return idx_voxel


def _get_mesh_stl(domain_stl):
    """
    Load meshes from STL files and find the minimum and maximum coordinates.
    Find the bounding box for all the meshes (minimum and maximum coordinates).
    """

    # init STL mesh list
    mesh_stl = []
    mesh_all = []

    # init the coordinate (minimum and maximum coordinates)
    xyz_min = np.full(3, +np.inf, dtype=np.float_)
    xyz_max = np.full(3, -np.inf, dtype=np.float_)

    # load the STL files and find the bounding box
    for tag, domain_stl_tmp in domain_stl.items():
        for domain_stl_tmp_tmp in domain_stl_tmp:
            # extract the data
            scale = domain_stl_tmp_tmp["scale"]
            offset = domain_stl_tmp_tmp["offset"]
            filename = domain_stl_tmp_tmp["filename"]

            # load the STL
            mesh = _get_load_stl(scale, offset, filename)

            # find the bounds
            (x_min, x_max, y_min, y_max, z_min, z_max) = mesh.bounds
            tmp_min = np.array((x_min, y_min, z_min), dtype=np.float_)
            tmp_max = np.array((x_max, y_max, z_max), dtype=np.float_)

            # update the bounds
            xyz_min = np.minimum(xyz_min, tmp_min)
            xyz_max = np.maximum(xyz_max, tmp_max)

            # assign the mesh
            mesh_all.append(mesh)
            mesh_stl.append({"tag": tag, "mesh": mesh})

    # merge all the meshes
    reference = pv.MultiBlock(mesh_all)
    reference = reference.combine().extract_surface()

    return mesh_stl, reference, xyz_min, xyz_max


def _get_voxel_size(d, xyz_max, xyz_min):
    """
    Get the parameters (size, dimension, and center) of the voxel structure.
    """

    # geometry size
    c = (xyz_max+xyz_min)/2

    # extract the number of voxels and the voxel size
    n = np.rint((xyz_max-xyz_min)/d)
    d = (xyz_max-xyz_min)/n

    # cast data
    d = d.astype(np.float_)
    n = n.astype(np.int_)

    # check voxel validity
    if not np.all(d > 0):
        RuntimeError("invalid voxel dimension: should be positive")
    # check voxel validity
    if not np.all(n > 0):
        RuntimeError("invalid voxel number: should be positive")

    return n, d, c


def _get_voxel_grid(n, d, c):
    """
    Construct a PyVista uniform grid for the complete voxel structure.
    The grid is located around the STL meshes to be voxelized.
    """

    # create a uniform grid for the complete structure
    grid = pv.ImageData()

    # set the array size and the voxel size
    grid.origin = c-(n*d)/2
    grid.dimensions = n+1
    grid.spacing = d

    # add indices for tracking the voxels after voxelization
    grid["idx"] = np.arange(np.prod(n), dtype=np.int_)

    # cast is required for voxelization
    grid = grid.cast_to_unstructured_grid()

    return grid


def _get_domain_def(grid, domain_stl, mesh_stl):
    """
    Voxelize meshes and assign the indices to a dict.
    """

    # init the domain dict
    domain_def = {}
    for tag in domain_stl:
        domain_def[tag] = np.empty(0, np.int_)

    # voxelize the meshes
    for mesh_stl_tmp in mesh_stl:
        # extract the data
        tag = mesh_stl_tmp["tag"]
        mesh = mesh_stl_tmp["mesh"]

        # voxelize and get the indices
        idx_voxel = _get_voxelize_stl(grid, mesh)

        # append the indices into the corresponding domain
        domain_def[tag] = np.append(domain_def[tag], idx_voxel)

    # remove duplicates
    for tag, idx_voxel in domain_def.items():
        idx_voxel = np.unique(idx_voxel)
        LOGGER.debug("%s: n_voxel = %d" % (tag, len(idx_voxel)))
        domain_def[tag] = idx_voxel

    return domain_def


def get_mesh(param, domain_stl):
    """
    Transform STL files into a 3D voxel structure.
    Each STL file corresponds to a domain of the 3D voxel structure.
    """

    # extract the data
    d = param["d"]
    xyz_min = param["xyz_min"]
    xyz_max = param["xyz_max"]

    # load the mesh and get the STL bounds
    LOGGER.debug("load STL files")
    (mesh_stl, reference, xyz_min_stl, xyz_max_stl) = _get_mesh_stl(domain_stl)

    # if provided, the specified bounds are used, otherwise the STL bounds are used
    if xyz_min is not None:
        xyz_min = np.array(xyz_min, np.float_)
    else:
        xyz_min = xyz_min_stl
    if xyz_max is not None:
        xyz_max = np.array(xyz_max, np.float_)
    else:
        xyz_max = xyz_max_stl

    # geometry size
    LOGGER.debug("get the voxel size")
    (n, d, c) = _get_voxel_size(d, xyz_max, xyz_min)

    # check total size
    nv = np.prod(n)
    if (VOXEL_TOTAL is not None) and (nv > VOXEL_TOTAL):
        raise RuntimeError("invalid size of the voxel structure: %d" % nv)

    # get the uniform grid
    LOGGER.debug("get the voxel grid")
    grid = _get_voxel_grid(n, d, c)

    # voxelize the meshes and get the indices
    LOGGER.debug("voxelize STL files")
    domain_def = _get_domain_def(grid, domain_stl, mesh_stl)

    # check used size
    nv = sum(len(idx) for idx in domain_def.values())
    if (VOXEL_USED is not None) and (nv > VOXEL_USED):
        raise RuntimeError("invalid number of used voxels: %d" % nv)

    # cast to lists
    n = n.tolist()
    d = d.tolist()
    c = c.tolist()

    # cast reference mesh
    reference = {
        "faces": np.array(reference.faces, dtype=np.int_),
        "points": np.array(reference.points, dtype=np.float_),
    }

    return n, d, c, domain_def, reference



"""
Module for reading and writing the test data.

Read the prescribed test results.
Write the prescribed test results.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os.path
import datetime
import json

# get the path the folder
PATH_ROOT = os.path.dirname(__file__)

# construct the test folder path
FOLDER_TESTS = os.path.join(PATH_ROOT, "..", "data")


def write_results(name, mesher, solver):
    """
    Write the file containing the prescribed test results.
    """

    # get timestamp
    timestamp = str(datetime.datetime.now())

    # get metadata
    metadata = {
        "name": name,
        "timestamp": timestamp,
    }

    # assemble results
    data_test = {"metadata": metadata, "mesher": mesher, "solver": solver}

    # file containing the test results
    file_test = os.path.join(FOLDER_TESTS, name + ".json")

    with open(file_test, "w") as fid:
        json.dump(data_test, fid, indent=4)


def read_results(name):
    """
    Load the file containing the prescribed test results.
    """

    # file containing the test results
    file_test = os.path.join(FOLDER_TESTS, name + ".json")

    # load the test results
    with open(file_test, "r") as fid:
        data_test = json.load(fid)

    # extract results
    mesher = data_test["mesher"]
    solver = data_test["solver"]

    return mesher, solver



"""
Compute and display metrics about 3D voxel structures.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec import log

# get a logger
LOGGER = log.get_logger("SUMMARY")


def get_status(n, d, s, c, domain_def, connection_def):
    """
    Get a dict summarizing a 3D voxel structure.
    """

    # extract the voxel data
    (nx, ny, nz) = n
    (dx, dy, dz) = d
    (cx, cy, cz) = c
    (sx, sy, sz) = s

    # compute voxel numbers
    n_total = nx*ny*nz
    n_graph = len(connection_def)
    n_domain = len(domain_def)

    # get the used voxels
    n_used = sum(len(idx) for idx in domain_def.values())

    # voxel utilization ratio
    ratio = n_used/n_total

    # get the volume and areas
    V_voxel = dx*dy*dz
    V_total = n_total*V_voxel
    V_used = n_used*V_voxel
    A_xy = sx*sy
    A_yz = sy*sz
    A_xz = sx*sz

    # assign data
    voxel_status = {
        "n_total": n_total,
        "n_used": n_used,
        "A_xy": A_xy,
        "A_yz": A_yz,
        "A_xz": A_xz,
        "V_total": V_total,
        "V_used": V_used,
        "ratio": ratio,
        "n_domain": n_domain,
        "n_graph": n_graph,
    }

    # display status
    LOGGER.debug("voxel size")
    with log.BlockIndent():
        LOGGER.debug("n = (%d, %d, %d)" % (nx, ny, nz))
        LOGGER.debug("d = (%.2e, %.2e, %.2e)" % (dx, dy, dz))
        LOGGER.debug("s = (%.2e, %.2e, %.2e)" % (sx, sy, sz))
        LOGGER.debug("c = (%.2e, %.2e, %.2e)" % (cx, cy, cz))
        LOGGER.debug("A = (%.2e, %.2e, %.2e)" % (A_xy, A_yz, A_xz))

    # plot the voxel number
    LOGGER.debug("voxel summary")
    with log.BlockIndent():
        LOGGER.debug("V_total = %.2e" % V_total)
        LOGGER.debug("V_used = %.2e" % V_used)
        LOGGER.debug("n_total = %d" % n_total)
        LOGGER.debug("n_used = %d" % n_used)
        LOGGER.debug("ratio = %.2e" % ratio)
        LOGGER.debug("n_domain = %d" % n_domain)
        LOGGER.debug("n_graph = %d" % n_graph)

    # plot the domain size
    LOGGER.debug("voxel domain")
    with log.BlockIndent():
        for tag, idx in domain_def.items():
            LOGGER.debug("%s = %d" % (tag, len(idx)))

    return voxel_status



"""
Module for checking the shape mesher data.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_check import datachecker


def _check_shape_data(shape_type, shape_data):
    """
    Check the validity of a single shape definition.
    """

    if shape_type == "pad":
        # check type and extract
        datachecker.check_dict("shape", shape_data, key_list=["coord", "diameter"])
        diameter = shape_data["diameter"]
        coord = shape_data["coord"]

        # check data
        datachecker.check_float("diameter", diameter, is_positive=True, can_be_zero=True)
        datachecker.check_float_pts("coord", coord, size=2, can_be_empty=False)
        datachecker.check_assert("coord", len(coord) >= 1, "pad coordinate should have at least one element")
    elif shape_type == "trace":
        # check type and extract
        datachecker.check_dict("shape", shape_data, key_list=["coord", "width"])
        width = shape_data["width"]
        coord = shape_data["coord"]

        # check data
        datachecker.check_float("width", width, is_positive=True, can_be_zero=True)
        datachecker.check_float_pts("coord", shape_data["coord"], size=2, can_be_empty=False)
        datachecker.check_assert("coord", len(coord) >= 2, "trace coordinate should have at least two elements")
    elif shape_type == "polygon":
        # check type and extract
        datachecker.check_dict("shape", shape_data, key_list=["coord_shell", "coord_holes", "buffer"])
        buffer = shape_data["buffer"]
        coord_shell = shape_data["coord_shell"]
        coord_holes = shape_data["coord_holes"]

        # check data
        datachecker.check_float("buffer", buffer, is_positive=True, can_be_zero=True, can_be_none=True)
        datachecker.check_float_pts("coord_shell", coord_shell, size=2, can_be_empty=False)
        datachecker.check_assert("coord_shell", len(coord_shell) >= 3, "polygon coordinate should have at least three elements")
        datachecker.check_list("coord_holes", coord_holes, can_be_empty=True)
        for coord in coord_holes:
            datachecker.check_float_pts("coord_holes", coord, size=2, can_be_empty=False)
            datachecker.check_assert("coord_holes", len(coord) >= 3, "polygon coordinate should have at least three elements")
    else:
        raise ValueError("invalid shape type")


def _check_geometry_shape(layer_list, geometry_shape):
    """
    Check the validity of the domain shape definition.
    """

    # check type
    datachecker.check_dict("geometry_shape", geometry_shape, can_be_empty=False)

    # check content
    for tag, geometry_shape_tmp in geometry_shape.items():
        datachecker.check_list("geometry_shape", geometry_shape_tmp, can_be_empty=False)
        for geometry_shape_tmp_tmp in geometry_shape_tmp:
            # check type
            key_list = ["shape_layer", "shape_operation", "shape_type", "shape_data"]
            datachecker.check_dict("geometry_shape", geometry_shape_tmp_tmp, key_list=key_list)

            # extract the data
            shape_layer = geometry_shape_tmp_tmp["shape_layer"]
            shape_operation = geometry_shape_tmp_tmp["shape_operation"]
            shape_type = geometry_shape_tmp_tmp["shape_type"]
            shape_data = geometry_shape_tmp_tmp["shape_data"]

            # check data
            datachecker.check_choice("shape_operation", shape_operation, ["add", "sub"])
            datachecker.check_choice("shape_type", shape_type, ["trace", "pad", "polygon"])
            datachecker.check_list("shape_layer", shape_layer, can_be_empty=False)
            datachecker.check_dict("shape_data", shape_data)

            # check layer
            for shape_layer_tmp in shape_layer:
                datachecker.check_choice("shape_layer", shape_layer_tmp, layer_list)

            # check data
            _check_shape_data(shape_type, shape_data)


def _check_layer_stack(layer_stack):
    """
    Check the validity of the layer stack definition.
    """

    # list with the layer name
    layer_list = []

    # check type
    datachecker.check_list("layer_stack", layer_stack, can_be_empty=False)

    # check value
    for layer_stack_tmp in layer_stack:
        # check type
        key_list = ["n_layer", "tag_layer"]
        datachecker.check_dict("layer_stack", layer_stack_tmp, key_list=key_list)

        # extract the data
        n_layer = layer_stack_tmp["n_layer"]
        tag_layer = layer_stack_tmp["tag_layer"]

        # check data
        datachecker.check_integer("n_layer", n_layer, is_positive=True, can_be_zero=False)
        datachecker.check_string("tag_layer", tag_layer, can_be_empty=False)

        # add layer name
        layer_list.append(tag_layer)

    return layer_list


def _check_param(param):
    """
    Check the voxel structure parameters.
    """

    # check type
    key_list = ["dx", "dy", "dz", "cz", "tol", "xy_min", "xy_max"]
    datachecker.check_dict("param", param, key_list=key_list)

    # check data
    datachecker.check_float("dx", param["dx"], is_positive=True, can_be_zero=False)
    datachecker.check_float("dy", param["dy"], is_positive=True, can_be_zero=False)
    datachecker.check_float("dz", param["dz"], is_positive=True, can_be_zero=False)
    datachecker.check_float("cz", param["cz"], is_positive=False, can_be_zero=True)
    datachecker.check_float("tol", param["tol"], is_positive=True, can_be_zero=False, can_be_none=True)
    datachecker.check_float_array("xy_min", param["xy_min"], size=2, is_positive=False, can_be_zero=True, can_be_none=True)
    datachecker.check_float_array("xy_max", param["xy_max"], size=2, is_positive=False, can_be_zero=True, can_be_none=True)


def check_data_voxelize(data_voxelize):
    """
    Check the data used for voxelization.
    """

    # check type
    key_list = ["param", "layer_stack", "geometry_shape"]
    datachecker.check_dict("data_voxelize", data_voxelize, key_list=key_list)

    # extract field
    param = data_voxelize["param"]
    layer_stack = data_voxelize["layer_stack"]
    geometry_shape = data_voxelize["geometry_shape"]

    # check voxel structure parameters
    _check_param(param)

    # check layer stack definition
    layer_list = _check_layer_stack(layer_stack)

    # check the domain shape definition
    _check_geometry_shape(layer_list, geometry_shape)

    # get the domain name
    domain_list = geometry_shape.keys()

    return domain_list



"""
Different functions for resolving conflict between domains.
Detect and remove shared indices (conflict) between domains.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np


def _get_solve_overlap(domain_def, domain_resolve, domain_keep):
    """
    Detect and remove shared indices (conflict) between two domains.
    The conflict is solved in the following way:
        - the reference domain remains unchanged
        - the shared indices are removed from the domain to be fixed
    """

    for domain_keep_tmp in domain_keep:
        for domain_resolve_tmp in domain_resolve:
            # get the indices
            idx_keep = domain_def[domain_keep_tmp]
            idx_resolve = domain_def[domain_resolve_tmp]

            # resolve the conflict
            idx_resolve = np.setdiff1d(idx_resolve, idx_keep)

            # update the domain indices
            domain_def[domain_resolve_tmp] = idx_resolve

    return domain_def


def _get_resolution(domain_def):
    """
    Check that all the conflicts are resolved.
    """

    # assemble all the indices
    idx_all = np.empty(0, dtype=np.int_)
    for idx in domain_def.values():
        idx_all = np.append(idx_all, idx)

    # check that all the conflicts are resolved
    if not (len(np.unique(idx_all)) == len(idx_all)):
        raise RuntimeError("invalid domain: domain indices should be unique")


def get_conflict(domain_def, domain_conflict):
    """
    Detect and remove shared indices (conflict) between domains.
    The direction of the conflict resolution (between two domains) is specified by the user.
    At the end, check that all shared indices have been removed.
    """

    # resolve the conflicts for all the specified domain pairs
    for domain_conflict_tmp in domain_conflict:
        # extract the data
        domain_resolve = domain_conflict_tmp["domain_resolve"]
        domain_keep = domain_conflict_tmp["domain_keep"]

        # solve the conflict
        domain_def = _get_solve_overlap(domain_def, domain_resolve, domain_keep)

    # check that the conflicts are resolved
    _get_resolution(domain_def)

    return domain_def



"""
Main script for plotting the solution of a PEEC problem.
Plot the following features:
    - material description
    - potential (electric and magnetic)
    - current density divergence and magnetic charges
    - current density and flux density
    - magnetic field
    - solver convergence

Three different mode are available for the plots:
    - windowed mode: with the Qt Framework
    - notebook mode: with Jupyter notebook
    - silent mode: nothing is shown and the program exit

The 3D plots are generated with PyVista.
The 2D plots are generated with Matplotlib.
The plots can be saved as PNG images.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_visualization import data_plotter
from pypeec.lib_visualization import data_voxel
from pypeec.lib_visualization import manage_pyvista
from pypeec.lib_visualization import manage_matplotlib
from pypeec.lib_visualization import manage_plotgui
from pypeec.lib_check import check_data_visualization
from pypeec.lib_check import check_data_options
from pypeec import log

# get a logger
LOGGER = log.get_logger("PLOTTER")


def _get_grid_voxel(data_init, data_sweep):
    """
    Convert the complete voxel geometry into a PyVista uniform grid.
    Convert the non-empty voxel geometry into a PyVista unstructured grid.
    Add the solver results to the grid.
    """

    # extract the data
    n = data_init["n"]
    d = data_init["d"]
    c = data_init["c"]
    idx_vc = data_init["idx_vc"]
    idx_vm = data_init["idx_vm"]
    idx_src_c = data_init["idx_src_c"]
    idx_src_v = data_init["idx_src_v"]
    pts_cloud = data_init["pts_cloud"]

    # extract the data
    V_vc = data_sweep["V_vc"]
    V_vm = data_sweep["V_vm"]
    J_vc = data_sweep["J_vc"]
    B_vm = data_sweep["B_vm"]
    S_vc = data_sweep["S_vc"]
    Q_vm = data_sweep["Q_vm"]
    P_vc = data_sweep["P_vc"]
    P_vm = data_sweep["P_vm"]
    H_pts = data_sweep["H_pts"]
    res = data_sweep["res"]
    conv = data_sweep["conv"]

    # get voxel indices
    idx = data_plotter.get_voxel(idx_vc, idx_vm)

    # convert the voxel geometry into PyVista grids
    grid = data_voxel.get_grid(n, d, c)
    voxel = data_voxel.get_voxel(grid, idx)
    point = data_voxel.get_point(pts_cloud)

    # add the problem solution to the grid
    voxel = data_plotter.set_voxel_material(voxel, idx, idx_vc, idx_vm, idx_src_c, idx_src_v)

    # set the electric variables
    voxel = data_plotter.set_voxel_scalar(voxel, idx, idx_vc, V_vc, "V_c")
    voxel = data_plotter.set_voxel_scalar(voxel, idx, idx_vc, S_vc, "S_c")
    voxel = data_plotter.set_voxel_scalar(voxel, idx, idx_vc, P_vc, "P_c")
    voxel = data_plotter.set_voxel_vector(voxel, idx, idx_vc, J_vc, "J_c")

    # set the magnetic variables
    voxel = data_plotter.set_voxel_scalar(voxel, idx, idx_vm, V_vm, "V_m")
    voxel = data_plotter.set_voxel_scalar(voxel, idx, idx_vm, Q_vm, "Q_m")
    voxel = data_plotter.set_voxel_scalar(voxel, idx, idx_vm, P_vm, "P_m")
    voxel = data_plotter.set_voxel_vector(voxel, idx, idx_vm, B_vm, "B_m")

    # add the magnetic field
    point = data_plotter.set_magnetic_field(point, H_pts)

    return grid, voxel, point, res, conv


def _get_plot(tag, data_plotter, grid, voxel, point, res, conv, gui_obj):
    """
    Make a plot with the specified user settings.
    The plot contains the following elements:
        - the geometry as wireframe
        - the axis, color bar, and legend
        - the payload (material description, scalar plots, or arrow plots)
    """

    # extract the data
    framework = data_plotter["framework"]
    title = data_plotter["title"]
    layout = data_plotter["layout"]
    data_window = data_plotter["data_window"]
    data_plot = data_plotter["data_plot"]
    data_options = data_plotter["data_options"]

    # make the plots
    if framework == "pyvista":
        # get the plotter (with the Qt framework)
        pl = gui_obj.open_pyvista(tag, title, data_window)

        # make the plot
        manage_pyvista.get_plot_plotter(pl, grid, voxel, point, layout, data_plot, data_options)
    elif framework == "matplotlib":
        # get the figure (with the Qt framework)
        fig = gui_obj.open_matplotlib(tag, title, data_window)

        # make the plot
        manage_matplotlib.get_plot_plotter(fig, res, conv, layout, data_plot, data_options)
    else:
        raise ValueError("invalid plot framework")


def _get_sweep(tag_sweep, data_sweep, data_init, data_plotter, gui_obj):
    """
    Parse the geometry and make the plots for a specified sweep.
    """

    # handle the data
    LOGGER.info("parse data")
    (grid, voxel, point, res, conv) = _get_grid_voxel(data_init, data_sweep)

    # make the plots
    with log.BlockTimer(LOGGER, "generate plots"):
        for i, (tag_plot, data_plotter_tmp) in enumerate(data_plotter.items()):
            LOGGER.info("plotting %d / %d / %s" % (i + 1, len(data_plotter), tag_plot))
            _get_plot(tag_sweep + "_" + tag_plot, data_plotter_tmp, grid, voxel, point, res, conv, gui_obj)


def run(
        data_solution, data_plotter,
        tag_sweep=None, tag_plot=None, plot_mode="qt", folder=".", name=None
):
    """
    Main script for plotting the solution of a PEEC problem.
    Handle invalid data with exceptions.
    """

    # check the solution data
    LOGGER.info("check the solution data")
    (data_init, data_sweep) = check_data_options.check_data_solution(data_solution)

    # check the input data
    LOGGER.info("check the input data")
    check_data_visualization.check_data_plotter(data_plotter)
    check_data_options.check_plot_options(plot_mode, folder, name)
    check_data_options.check_tag_list(data_sweep, tag_sweep)
    check_data_options.check_tag_list(data_plotter, tag_plot)

    # find the plots
    if tag_sweep is not None:
        data_sweep = {key: data_sweep[key] for key in tag_sweep}
    if tag_plot is not None:
        data_plotter = {key: data_plotter[key] for key in tag_plot}

    # create the Qt app (should be at the beginning)
    LOGGER.info("init the plot manager")
    gui_obj = manage_plotgui.PlotGui(plot_mode, folder, name)

    # plot the sweeps
    for tag_sweep, data_sweep_tmp in data_sweep.items():
        with log.BlockTimer(LOGGER, "plot sweep: " + tag_sweep):
            _get_sweep(tag_sweep, data_sweep_tmp, data_init, data_plotter, gui_obj)

    # enter the event loop (should be at the end, blocking call)
    gui_obj.show()



"""
Module for running the sweeps with interdependencies.
Interdependency means that the result of a sweep can be used as the input for another one.

Build a tree representing the interdependencies between the sweeps.
Check that the interdependencies are not impossible (no cyclical dependencies).
Run the sweeps in the correct order and return the results.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import joblib
from pypeec import config
from pypeec import log

# get config
SWEEP_POOL = config.SWEEP_POOL


def _get_tree_tag(sweep_config, tag):
    """
    Find the run configurations for a defined init tag.
    """

    tag_sub = []
    for tag_run, tag_init in sweep_config.items():
        if tag_init == tag:
            tag_sub.append(tag_run)

    return tag_sub


def _get_tree_create(sweep_config):
    """
    Represent the sweep interdependencies with a tree.
    The root of the tree is the starting point for the computing the sweeps.
    """

    # dict for storing the tree
    sweep_tree = {}

    # add the dependencies between the sweeps
    for tag_tmp in sweep_config.keys():
        sweep_tree[tag_tmp] = _get_tree_tag(sweep_config, tag_tmp)

    # the root node is linking the sweep without dependencies
    sweep_tree[None] = _get_tree_tag(sweep_config, None)

    return sweep_tree


def _get_tree_check(sweep_tree, tag_init, init_list):
    """
    Check that the dependencies between the sweeps are solvable.
    Run through the tree from the root and gather the nodes.
    """

    # find the sweeps to be computed
    tag_sub = sweep_tree[tag_init]

    # add the design to the computed design
    init_list += tag_sub

    # recursive call for the dependent sweeps
    for tag_tmp in tag_sub:
        init_list = _get_tree_check(sweep_tree, tag_tmp, init_list)

    return init_list


def _get_parallel_loop(fct_compute, arg_list):
    """
    Run a loop (serial or parallel).
    """

    if SWEEP_POOL is None:
        out_list = []
        for arg in arg_list:
            out = fct_compute(*arg)
            out_list.append(out)
    else:
        # get the log global parameters
        global_warning = log.get_warning()
        (global_timestamp, global_level) = log.get_global()

        # wrap the compute function for setting globals
        def fct_joblib(*args):
            log.set_warning(global_warning)
            log.set_global(global_timestamp, global_level)
            out = fct_compute(*args)
            warning = log.get_warning()
            return out, warning

        # run the parallel loop
        pool_list = joblib.Parallel(n_jobs=SWEEP_POOL, backend="loky")(
            joblib.delayed(fct_joblib)(*arg) for arg in arg_list
        )

        # unpack the warnings
        (out_list, warning_list) = zip(*pool_list)

        # set the warnings
        global_warning = global_warning or all(warning_list)
        log.set_warning(global_warning)

    return out_list


def _get_tree_compute(sweep_tree, sweep_param, fct_compute, tag_init, output, init):
    """
    Compute the sweeps with the dependencies.
    This is done by walking through the graph from the root.
    """

    # find the sweeps to be computed
    tag_sub = sweep_tree[tag_init]

    # find the dependency for the sweeps
    if tag_init is None:
        init_tmp = None
    else:
        init_tmp = init[tag_init]

    # get the input
    data_sub = [sweep_param[tag_tmp] for tag_tmp in tag_sub]
    init_sub = [init_tmp]*len(tag_sub)

    # return if there is nothing to compute
    if not tag_sub:
        return output, init

    # run the serial or parallel loop
    arg_list = zip(tag_sub, data_sub, init_sub)
    out_list = _get_parallel_loop(fct_compute, arg_list)

    # assemble the results
    for tag_tmp, out_tmp in zip(tag_sub, out_list):
        (output_tmp, sol_tmp) = out_tmp
        output[tag_tmp] = output_tmp
        init[tag_tmp] = sol_tmp

    # recursive call for the dependent sweeps
    for tag_tmp in tag_sub:
        (output, init) = _get_tree_compute(sweep_tree, sweep_param, fct_compute, tag_tmp, output, init)

    return output, init


def get_run_sweep(sweep_config, sweep_param, fct_compute):
    """
    Build a tree representing the interdependencies between the sweeps.
    Check that the interdependencies are not impossible (no cyclical dependencies).
    Run the sweeps in the correct order and return the results.
    """

    # create a representing the interdependencies between the sweeps
    sweep_tree = _get_tree_create(sweep_config)

    # ensure that the interdependencies are solvable
    init_list = _get_tree_check(sweep_tree, None, [])
    if len(init_list) != len(sweep_config):
        raise RuntimeError("invalid sweep: cannot solve the sweep dependencies")

    # init the dict for the output data
    output = {}

    # init the dict for the interdependent data
    init = {}

    # compute the sweep in the correct order (starting from the tree root)
    (output, init) = _get_tree_compute(sweep_tree, sweep_param, fct_compute, None, output, init)

    return output



"""
Module containing the command line script.
    - Used for calling the mesher, solver, viewer, and plotter.
    - Parse the command line arguments.
    - Call the corresponding entry point.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import sys
import argparse
from pypeec import main
from pypeec import config


def _get_parser():
    """
    Create a command line parser with a description.
    """

    # get the parser
    parser = argparse.ArgumentParser(
        prog="pypeec",
        description="PyPEEC - 3D PEEC Solver",
        epilog="Thomas Guillod - Dartmouth College",
        allow_abbrev=False,
    )

    # get version
    version = config.get_version()

    # display the version
    parser.add_argument(
        "-v", "--version",
        action="version",
        version="PyPEEC %s" % version,
    )

    # add subparsers
    subparsers = parser.add_subparsers(
        required=True,
        dest="command",
        title="command",
    )

    return parser, subparsers


def _get_arg_mesher(subparsers):
    """
    Add the mesher arguments.
    """

    # add the subparser
    parser = subparsers.add_parser(
        "mesher",
        aliases=["me"],
        help="mesh a geometry into a 3D voxel structure",
    )

    # add the arguments
    parser.add_argument(
        "-ge", "--geometry",
        help="geometry file (input / JSON or YAML)",
        required=True,
        metavar="file",
        dest="file_geometry",
    )
    parser.add_argument(
        "-vo", "--voxel",
        help="voxel file (output / pickle)",
        required=True,
        metavar="file",
        dest="file_voxel",
    )
    parser.add_argument(
        "-t", "--truncated",
        help="truncate the results (default: full results)",
        action="store_true",
        dest="is_truncated",
    )


def _get_arg_viewer(subparsers):
    """
    Add the viewer arguments.
    """

    # add the subparser
    parser = subparsers.add_parser(
        "viewer",
        aliases=["vi"],
        help="visualization of a 3D voxel structure",
    )

    # add the arguments
    parser.add_argument(
        "-vo", "--voxel",
        help="voxel file (input / pickle)",
        required=True,
        metavar="file",
        dest="file_voxel",
    )
    parser.add_argument(
        "-vi", "--viewer",
        help="viewer file (input / JSON or YAML)",
        required=True,
        metavar="file",
        dest="file_viewer",
    )
    parser.add_argument(
        "-tp", "--tag_plot",
        help="list of plots to be shown (default: all the plots)",
        nargs='+',
        default=None,
        metavar="tag_plot",
        dest="tag_plot",
    )
    parser.add_argument(
        "-pm", "--plot_mode",
        help="selection of the plot mode (default: window)",
        choices=["qt", "nb", "save", "none"],
        default="qt",
        dest="plot_mode",
    )
    parser.add_argument(
        "-f", "--folder",
        help="folder for saving the screenshots (default: cwd)",
        default=".",
        metavar="folder",
        dest="folder",
    )


def _get_arg_solver(subparsers):
    """
    Add the solver arguments.
    """

    # add the subparser
    parser = subparsers.add_parser(
        "solver",
        aliases=["so"],
        help="solve a problem with the PEEC method",
    )

    # add the arguments
    parser.add_argument(
        "-vo", "--voxel",
        help="voxel file (input / pickle)",
        required=True,
        metavar="file",
        dest="file_voxel",
    )
    parser.add_argument(
        "-pr", "--problem",
        help="problem file (input / JSON or YAML)",
        required=True,
        metavar="file",
        dest="file_problem",
    )
    parser.add_argument(
        "-to", "--tolerance",
        help="tolerance file (input / JSON or YAML)",
        required=True,
        metavar="file",
        dest="file_tolerance",
    )
    parser.add_argument(
        "-so", "--solution",
        help="solution file (output / pickle)",
        required=True,
        metavar="file",
        dest="file_solution",
    )
    parser.add_argument(
        "-t", "--truncated",
        help="truncate the results (default: full results)",
        action="store_true",
        dest="is_truncated",
    )


def _get_arg_plotter(subparsers):
    """
    Add the plotter arguments.
    """

    # add the subparser
    parser = subparsers.add_parser(
        "plotter",
        aliases=["pl"],
        help="plot the solution of a PEEC problem",
    )

    # add the arguments
    parser.add_argument(
        "-so", "--solution",
        help="solution file (input / pickle)",
        required=True,
        metavar="file",
        dest="file_solution",
    )
    parser.add_argument(
        "-pl", "--plotter",
        help="plotter file (input / JSON or YAML)",
        required=True,
        metavar="file",
        dest="file_plotter",
    )
    parser.add_argument(
        "-ts", "--tag_sweep",
        help="list of sweeps to be shown (default: show the sweeps)",
        nargs='+',
        default=None,
        metavar="tag_sweep",
        dest="tag_sweep",
    )
    parser.add_argument(
        "-tp", "--tag_plot",
        help="list of plots to be shown (default: all the plots)",
        nargs='+',
        default=None,
        metavar="tag_plot",
        dest="tag_plot",
    )
    parser.add_argument(
        "-pm", "--plot_mode",
        help="selection of the plot mode (default: window)",
        choices=["qt", "nb", "save", "none"],
        default="qt",
        dest="plot_mode",
    )
    parser.add_argument(
        "-f", "--folder",
        help="folder for saving the screenshots (default: cwd)",
        default=".",
        metavar="folder",
        dest="folder",
    )


def _get_arg_extract(subparsers):
    """
    Add the data extraction arguments (examples and documentation).
    """

    # add the examples parser
    parser = subparsers.add_parser(
        "examples",
        help="extract the examples",
    )
    parser.add_argument(
        help="path where the examples should be extracted",
        type=str,
        metavar="path",
        dest="path_extract",
    )

    # add the documentation parser
    parser = subparsers.add_parser(
        "documentation",
        help="extract the documentation",
    )
    parser.add_argument(
        help="path where the documentation should be extracted",
        type=str,
        metavar="path",
        dest="path_extract",
    )

    # add the config parser
    parser = subparsers.add_parser(
        "config",
        help="extract the default config file",
    )
    parser.add_argument(
        help="path where the config file should be extracted",
        type=str,
        metavar="path",
        dest="path_extract",
    )


def run_arguments(argv):
    """
    User script for running PyPEEC with given arguments.
        - The script offers a CLI for the mesher, solver, viewer, and plotter.
        - The script can also be used to extract the examples, documentation and config.

    Parameters
    ----------
    argv : list
        List with the command line arguments.

    Returns
    -------
    status : int
        The status exit code of the script.

    """

    # get the parser
    (parser, subparsers) = _get_parser()

    # add the sub-command arguments
    _get_arg_mesher(subparsers)
    _get_arg_viewer(subparsers)
    _get_arg_solver(subparsers)
    _get_arg_plotter(subparsers)
    _get_arg_extract(subparsers)

    # parse the config and get arguments
    try:
        args = parser.parse_args(argv)
    except SystemExit as status:
        return status.code

    # run the code
    try:
        if args.command in ["mesher", "me"]:
            main.run_mesher_file(
                args.file_geometry,
                args.file_voxel,
                is_truncated=args.is_truncated,
            )
        elif args.command in ["viewer", "vi"]:
            main.run_viewer_file(
                args.file_voxel,
                args.file_viewer,
                tag_plot=args.tag_plot,
                plot_mode=args.plot_mode,
                folder=args.folder,
            )
        elif args.command in ["solver", "so"]:
            main.run_solver_file(
                args.file_voxel,
                args.file_problem,
                args.file_tolerance,
                args.file_solution,
                is_truncated=args.is_truncated,
            )
        elif args.command in ["plotter", "pl"]:
            main.run_plotter_file(
                args.file_solution,
                args.file_plotter,
                tag_sweep=args.tag_sweep,
                tag_plot=args.tag_plot,
                plot_mode=args.plot_mode,
                folder=args.folder,
            )
        elif args.command == "examples":
            main.run_extract("examples.zip", True, args.path_extract)
        elif args.command == "documentation":
            main.run_extract("documentation.zip", True, args.path_extract)
        elif args.command == "config":
            main.run_extract("config.yaml", False, args.path_extract)
        else:
            raise ValueError("invalid command")
    except Exception:
        return 1
    else:
        return 0


def run_script():
    """
    User script for running PyPEEC with the command line arguments.
        - The script offers a CLI for the mesher, solver, viewer, and plotter.
        - The script can also be used to extract the examples, documentation and config.

    The script is installed with the package.
    The name of the command line script is "pypeec".
    """

    # get arguments
    argv = sys.argv[1:]

    # run the script
    status = run_arguments(argv)

    # exit code
    sys.exit(status)


if __name__ == "__main__":
    run_script()



"""
Test the tutorial example.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from tests.code import test_workflow
from pypeec import main


# duplicate of the test class
class TestTutorial(test_workflow.TestWorkflow):
    pass


# show the logo
main.run_hide_logo()

# add the tests
test_workflow.set_test(TestTutorial, "tutorial_api", "tutorial", False)
test_workflow.set_test(TestTutorial, "tutorial_script", "tutorial", True)



"""
Module for parsing voxel indices.
Transform the voxel indices into NumPy arrays.
Check the validity of the indices.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
from pypeec import log
from pypeec import config

# get a logger
LOGGER = log.get_logger("VOXEL")

# get problem size limits
VOXEL_TOTAL = config.PROBLEM_MAX_SIZE.VOXEL_TOTAL
VOXEL_USED = config.PROBLEM_MAX_SIZE.VOXEL_USED


def get_mesh(param, domain_index):
    """
    Check the domain definition (mapping between domain names and indices).
    """

    # extract the data
    n = param["n"]
    d = param["d"]
    c = param["c"]

    # no reference geometry, direct voxelization
    reference = None

    # check total size
    nv = np.prod(n)
    if (VOXEL_TOTAL is not None) and (nv > VOXEL_TOTAL):
        raise RuntimeError("invalid size of the voxel structure: %d" % nv)

    # init new domain indices
    domain_def = {}

    # check data
    for tag, idx in domain_index.items():
        # disp
        LOGGER.debug("%s: size = %d" % (tag, len(idx)))

        # parse the array
        idx_tmp = np.array(idx, dtype=np.int_)

        # check the indices
        if not (len(np.unique(idx_tmp)) == len(idx_tmp)):
            raise RuntimeError("invalid index: %s: indices should be unique" % tag)
        if not (np.all(idx_tmp >= 0) and np.all(idx_tmp < nv)):
            raise RuntimeError("invalid index: %s: invalid index range" % tag)

        # add the new item
        domain_def[tag] = idx_tmp

    # check used size
    nv = sum(len(idx) for idx in domain_def.values())
    if (VOXEL_USED is not None) and (nv > VOXEL_USED):
        raise RuntimeError("invalid number of used voxels: %d" % nv)

    return n, d, c, domain_def, reference



"""
Module for doing matrix-vector multiplication:
    - direct matrix multiplication
    - multiplication with FFT and circulant tensors

Three different types of matrices are supported:
    - potential: tensor representing a simple potential matrix
        - size of the last dimension of the input tensor = 1
        - number of dimensions of the input vector = 1
        - number of dimensions of the output vector = 1
    - inductance: tensor representing a block diagonal inductance matrix
        - size of the last dimension of the input tensor = 1
        - number of dimensions of the input vector = 3
        - number of dimensions of the output vector = 3
    - coupling: tensor representing a block off-diagonal coupling matrix
        - size of the last dimension of the input tensor = 3
        - number of dimensions of the input vector = 3
        - number of dimensions of the output vector = 3

A matrix-vector operator is returned for performing the matrix-vector multiplication.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_matrix import multiply_fft
from pypeec.lib_matrix import multiply_direct


def _get_multiply(data, vec_in, mult_type, flip):
    """
    Make a matrix-vector multiplication.
    """

    if mult_type == "fft":
        res_out = multiply_fft.get_multiply(data, vec_in, flip)
    elif mult_type == "direct":
        res_out = multiply_direct.get_multiply(data, vec_in, flip)
    else:
        raise ValueError("invalid multiplication library")

    return res_out


def _get_prepare(name, idx_out, idx_in, mat, mult_type):
    """
    Prepare the matrix for the multiplication.
    """

    if mult_type == "fft":
        data = multiply_fft.get_prepare(name, idx_out, idx_in, mat)
    elif mult_type == "direct":
        data = multiply_direct.get_prepare(name, idx_out, idx_in, mat)
    else:
        raise ValueError("invalid multiplication library")

    return data


def get_operator_potential(idx, mat, mult_type):
    """
    Get the linear matrix-vector operator for a simple potential matrix.
    """

    # prepare the matrix
    data = _get_prepare("potential", idx, idx, mat, mult_type)

    # function describing the matrix-vector multiplication
    def op(vec_in):
        res_out = _get_multiply(data, vec_in, mult_type, False)
        return res_out

    return op


def get_operator_inductance(idx, mat, mult_type):
    """
    Get the linear matrix-vector operator for a block diagonal inductance matrix.
    """

    # prepare the matrix
    data = _get_prepare("inductance", idx, idx, mat, mult_type)

    # function describing the matrix-vector multiplication
    def op(vec_in):
        res_out = _get_multiply(data, vec_in, mult_type, False)
        return res_out

    return op


def get_operator_coupling(idx_out, idx_in, mat, mult_type):
    """
    Get the linear matrix-vector operator for a block off-diagonal coupling matrix.
    """

    # prepare the matrix
    data = _get_prepare("coupling", idx_out, idx_in, mat, mult_type)

    # function describing the matrix-vector multiplication
    def op_for(vec_in):
        res_out = _get_multiply(data, vec_in, mult_type, False)
        return res_out

    # function describing the matrix-vector multiplication
    def op_rev(vec_in):
        res_out = _get_multiply(data, vec_in, mult_type, True)
        return res_out

    return op_for, op_rev



"""
Module containing the program entry points.
    - Used for calling the mesher, solver, viewer, and plotter.
    - Files can be used for the input and output data.
    - Data structure can be used for the input and output data.
    - The import statements for the different modules are located inside the code.
    - This allows for a minimization of the loaded dependencies.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import sys
import shutil
import importlib.resources
from pypeec import config
from pypeec import log
from pypeec import io

# create the logger
LOGGER = log.get_logger("MAIN")

# get the logo display status
DISPLAY_LOGO = config.DISPLAY_LOGO

# init the logo display status
STATUS_LOGO = False


def run_display_logo():
    """
    Display the logo as a splash screen
    """

    # variable with the logo status
    global STATUS_LOGO

    # display the logo
    if not STATUS_LOGO:
        try:
            with importlib.resources.open_text("pypeec.data", "pypeec.txt") as file_logo:
                data = file_logo.read()
                print("", flush=True, file=sys.stderr)
                print(data, flush=True, file=sys.stderr)
                print("", flush=True, file=sys.stderr)
        except UnicodeError:
            pass

    # logo has been displayed
    STATUS_LOGO = True


def run_hide_logo():
    """
    Prevent the display of the splash screen.
    """

    # variable with the logo status
    global STATUS_LOGO

    # logo should not be displayed
    STATUS_LOGO = True


def run_extract(data_name, is_zip, path_extract):
    """
    Extract data (config, examples, or documentation).

    Parameters
    ----------
    data_name : string
        Name of the file containing the data.
    is_zip : boolean
        Indicate if the data is a zip archive.
    path_extract : string
        Path where the data will be extracted.
    """

    # display logo
    if DISPLAY_LOGO:
        run_display_logo()

    # execute workflow
    try:
        LOGGER.info("data extraction")
        with importlib.resources.path("pypeec.data", data_name) as file_data:
            if is_zip:
                shutil.unpack_archive(file_data, path_extract)
            else:
                shutil.copy(file_data, path_extract)
    except Exception as ex:
        log.log_exception(LOGGER, ex)
        LOGGER.error("invalid termination")
        raise ex
    else:
        LOGGER.info("successful termination")


def run_mesher_data(data_geometry, **kwargs):
    """
    Main script for meshing the geometry and generating a 3D voxel structure.
        - Get the geometry data as an argument.
        - Return the created voxel data.

    Parameters
    ----------
    data_geometry : data
        The dict describes the geometry, meshing and resampling process.
    is_truncated : boolean
        If false, the complete results are returned (default).
        If true, the results are truncated to save space.
        This argument is optional.

    Returns
    -------
    data_voxel : data
        The dict describes the voxel structure.
    """

    # display logo
    if DISPLAY_LOGO:
        run_display_logo()

    # execute workflow
    try:
        # reset warning
        log.reset_warning()

        # load the tool
        LOGGER.info("load the mesher")
        from pypeec.run import mesher

        # run the tool
        LOGGER.info("run the mesher")
        data_voxel = mesher.run(data_geometry, **kwargs)
    except Exception as ex:
        log.log_exception(LOGGER, ex)
        LOGGER.error("invalid termination")
        raise ex
    else:
        if log.get_warning():
            LOGGER.warning("termination with warnings")
        else:
            LOGGER.info("successful termination")

    return data_voxel


def run_mesher_file(file_geometry, file_voxel, **kwargs):
    """
    Main script for meshing the geometry and generating a 3D voxel structure.
        - Load the geometry data from a file.
        - Write the created voxel data in a file.

    Parameters
    ----------
    file_geometry : filename
        The file content describes the geometry, meshing and resampling process.
        This input file is loaded by this function (JSON or YAML format).
    file_voxel : filename
        The file content describes the voxel structure.
        This output file is written by this function (JSON or Pickle format).
    is_truncated : boolean
        If false, the complete results are returned (default).
        If true, the results are truncated to save space.
        This argument is optional.
    """

    # display logo
    if DISPLAY_LOGO:
        run_display_logo()

    # load data
    try:
        LOGGER.info("load the input data")
        data_geometry = io.load_input(file_geometry)
    except Exception as ex:
        log.log_exception(LOGGER, ex)
        raise ex

    # run the tool
    data_voxel = run_mesher_data(data_geometry, **kwargs)

    # save results
    try:
        LOGGER.info("save the results")
        io.write_data(file_voxel, data_voxel)
    except Exception as ex:
        log.log_exception(LOGGER, ex)
        raise ex


def run_viewer_data(data_voxel, data_viewer, **kwargs):
    """
    Main script for visualizing a 3D voxel structure.
        - Get the voxel data as an argument.
        - Get the point data as an argument.
        - Get the viewer data as an argument.

    Parameters
    ----------
    data_voxel : data
        The dict describes the voxel structure.
    data_viewer: data
        The dict describes the different plots to be created.
    tag_plot : list
        The list describes plots to be shown.
        If None, all the plots are shown (default).
        This argument is optional.
    plot_mode : string
        If "qt", the Qt framework is used for the rendering (default).
        If "nb", the plots are rendered within the Jupyter notebook.
        If "save", the plots are not shown but saved as screenshots.
        If "none", the plots are not shown (test mode).
        This argument is optional.
    folder : string
        Folder name for saving the screenshots.
        The current directory is used as the default directory.
        This argument is optional.
    name : string
        Prepended at the beginning of the screenshot filenames.
        This argument is optional.
    """

    # display logo
    if DISPLAY_LOGO:
        run_display_logo()

    # execute workflow
    try:
        # reset warning
        log.reset_warning()

        # load the tool
        LOGGER.info("load the viewer")
        from pypeec.run import viewer

        # run the tool
        LOGGER.info("run the viewer")
        viewer.run(data_voxel, data_viewer, **kwargs)
    except Exception as ex:
        log.log_exception(LOGGER, ex)
        LOGGER.error("invalid termination")
        raise ex
    else:
        if log.get_warning():
            LOGGER.warning("termination with warnings")
        else:
            LOGGER.info("successful termination")


def run_viewer_file(file_voxel, file_viewer, **kwargs):
    """
    Main script for visualizing a 3D voxel structure.
        - Load the voxel data from a file.
        - Load the point data from a file.
        - Load the viewer data from a file.

    Parameters
    ----------
    file_voxel : filename
        The file content describes the voxel structure.
        This input file is loaded by this function (JSON or Pickle format).
    file_viewer: filename
        The file content describes the different plots to be created.
        This input file is loaded by this function (JSON or YAML format).
    tag_plot : list
        The list describes plots to be shown.
        If None, all the plots are shown (default).
        This argument is optional.
    plot_mode : string
        If "qt", the Qt framework is used for the rendering (default).
        If "nb", the plots are rendered within the Jupyter notebook.
        If "save", the plots are not shown but saved as screenshots.
        If "none", the plots are not shown (test mode).
        This argument is optional.
    folder : string
        Folder name for saving the screenshots.
        The current directory is used as the default directory.
        This argument is optional.
    name : string
        Prepended at the beginning of the screenshot filenames.
        This argument is optional.
    """

    # display logo
    if DISPLAY_LOGO:
        run_display_logo()

    # load data
    try:
        LOGGER.info("load the input data")
        data_voxel = io.load_data(file_voxel)
        data_viewer = io.load_input(file_viewer)
    except Exception as ex:
        log.log_exception(LOGGER, ex)
        raise ex

    # run the tool
    run_viewer_data(data_voxel, data_viewer, **kwargs)


def run_solver_data(data_voxel, data_problem, data_tolerance, **kwargs):
    """
    Main script for solving a problem with the PEEC solver.
        - Get the voxel data as an argument.
        - Get the problem data as an argument.
        - Get the tolerance data as an argument.
        - Return the created solution data.

    Parameters
    ----------
    data_voxel : data
        The dict describes the voxel structure.
    data_problem: data
        The dict describes the problem to be solved.
    data_tolerance: data
        The dict describes the numerical options.
    is_truncated : boolean
        If false, the complete results are returned (default).
        If true, the results are truncated to save space.
        This argument is optional.

    Returns
    -------
    data_solution : data
        The dict describes the problem solution.
    """

    # display logo
    if DISPLAY_LOGO:
        run_display_logo()

    # execute workflow
    try:
        # reset warning
        log.reset_warning()

        # load the tool
        LOGGER.info("load the solver")
        from pypeec.run import solver

        # run the tool
        LOGGER.info("run the solver")
        data_solution = solver.run(data_voxel, data_problem, data_tolerance, **kwargs)
    except Exception as ex:
        log.log_exception(LOGGER, ex)
        LOGGER.error("invalid termination")
        raise ex
    else:
        if log.get_warning():
            LOGGER.warning("termination with warnings")
        else:
            LOGGER.info("successful termination")

    return data_solution


def run_solver_file(file_voxel, file_problem, file_tolerance, file_solution, **kwargs):
    """
    Main script for solving a problem with the PEEC solver.
        - Load the voxel data from a file.
        - Load the problem data from a file.
        - Load the tolerance data from a file.
        - Write the created solution data in a file.

    Parameters
    ----------
    file_voxel : filename
        The file content describes the voxel structure.
        This input file is loaded by this function (JSON or Pickle format).
    file_problem: filename
        The file content describes the problem to be solved.
        This input file is loaded by this function (JSON or YAML format).
    file_tolerance: filename
        The file content describes the numerical options.
        This input file is loaded by this function (JSON or YAML format).
    file_solution : filename
        The file content describes the problem solution.
        This output file is written by this function (JSON or Pickle format).
    is_truncated : boolean
        If false, the complete results are returned (default).
        If true, the results are truncated to save space.
        This argument is optional.
    """

    # display logo
    if DISPLAY_LOGO:
        run_display_logo()

    # load data
    try:
        LOGGER.info("load the input data")
        data_voxel = io.load_data(file_voxel)
        data_problem = io.load_input(file_problem)
        data_tolerance = io.load_input(file_tolerance)
    except Exception as ex:
        log.log_exception(LOGGER, ex)
        raise ex

    # run the tool
    data_solution = run_solver_data(data_voxel, data_problem, data_tolerance, **kwargs)

    # save results
    try:
        LOGGER.info("save the results")
        io.write_data(file_solution, data_solution)
    except Exception as ex:
        log.log_exception(LOGGER, ex)
        raise ex


def run_plotter_data(data_solution, data_plotter, **kwargs):
    """
    Main script for plotting the solution of a PEEC problem.
        - Get the solution data as an argument.
        - Get the point data as an argument.
        - Get the plotter data as an argument.

    Parameters
    ----------
    data_solution : data
        The dict describes the problem solution.
    data_plotter : data
        The dict describes the different plots to be created.
    tag_sweep : list
        The list describes sweeps to be shown.
        If None, all the sweeps are shown (default).
        This argument is optional.
    tag_plot : list
        The list describes plots to be shown.
        If None, all the plots are shown (default).
        This argument is optional.
    plot_mode : string
        If "qt", the Qt framework is used for the rendering (default).
        If "nb", the plots are rendered within the Jupyter notebook.
        If "save", the plots are not shown but saved as screenshots.
        If "none", the plots are not shown (test mode).
        This argument is optional.
    folder : string
        Folder name for saving the screenshots.
        The current directory is used as the default directory.
        This argument is optional.
    name : string
        Prepended at the beginning of the screenshot filenames.
        This argument is optional.
    """

    # display logo
    if DISPLAY_LOGO:
        run_display_logo()

    # execute workflow
    try:
        # reset warning
        log.reset_warning()

        # load the tool
        LOGGER.info("load the plotter")
        from pypeec.run import plotter

        # run the tool
        LOGGER.info("run the plotter")
        plotter.run(data_solution, data_plotter, **kwargs)
    except Exception as ex:
        log.log_exception(LOGGER, ex)
        LOGGER.error("invalid termination")
        raise ex
    else:
        if log.get_warning():
            LOGGER.warning("termination with warnings")
        else:
            LOGGER.info("successful termination")


def run_plotter_file(file_solution, file_plotter, **kwargs):
    """
    Main script for plotting the solution of a PEEC problem.
        - Load the solution data from a file.
        - Load the point data from a file.
        - Load the plotter data from a file.

    Parameters
    ----------
    file_solution : filename
        The dict describes the problem solution.
        This input file is loaded by this function (JSON or Pickle format).
    file_plotter : filename
        The dict describes the different plots to be created.
        This input file is loaded by this function (JSON or YAML format).
    tag_sweep : list
        The list describes sweeps to be shown.
        If None, all the sweeps are shown (default).
        This argument is optional.
    tag_plot : list
        The list describes plots to be shown.
        If None, all the plots are shown (default).
        This argument is optional.
    plot_mode : string
        If "qt", the Qt framework is used for the rendering (default).
        If "nb", the plots are rendered within the Jupyter notebook.
        If "save", the plots are not shown but saved as screenshots.
        If "none", the plots are not shown (test mode).
        This argument is optional.
    folder : string
        Folder name for saving the screenshots.
        The current directory is used as the default directory.
        This argument is optional.
    name : string
        Prepended at the beginning of the screenshot filenames.
        This argument is optional.
    """

    # display logo
    if DISPLAY_LOGO:
        run_display_logo()

    # load data
    try:
        # load data
        LOGGER.info("load the input data")
        data_solution = io.load_data(file_solution)
        data_plotter = io.load_input(file_plotter)
    except Exception as ex:
        log.log_exception(LOGGER, ex)
        raise ex

    # run the tool
    run_plotter_data(data_solution, data_plotter, **kwargs)



"""
Module for transforming a series of 2D vector shapes into a 3D voxel structure.
The 2D geometry are stacked in order to create a voxel structure.

The following axis definition is used:
    - x: x-axis of the 2D shapes
    - y: y-axis of the 2D shapes
    - z: stacking dimension of the 2D geometries

The shape handling is done with Shapely.
The raster conversion is done with Rasterio.

Todo
----
- The warning triggered by Shapely and Rasterio should be handled in a cleaner way.
    - Rasterio triggers warnings for some coordinate transformations.
    - Shapely triggers warnings on Apple Silicon CPUs.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import warnings
import numpy as np
import pyvista as pv
import shapely as sha
import rasterio.features as raf
import rasterio.transform as rat
from pypeec import log
from pypeec import config

# prevent problematic linear transform to trigger warnings
warnings.filterwarnings("ignore", module="shapely")
warnings.filterwarnings("ignore", module="rasterio.features")
warnings.filterwarnings("ignore", module="rasterio.transform")

# get a logger
LOGGER = log.get_logger("SHAPE")

# get problem size limits
VOXEL_TOTAL = config.PROBLEM_MAX_SIZE.VOXEL_TOTAL
VOXEL_USED = config.PROBLEM_MAX_SIZE.VOXEL_USED


def _get_boundary_polygon(bnd, z_min):
    """
    Convert a Shapely boundary into a PyVista polygon.
    """

    # check that the boundary is closed
    if (not bnd.is_valid) or (not bnd.is_closed):
        raise RuntimeError("invalid shape: boundary is ill-formed")

    # get the 2D boundary
    xy = np.array(bnd.xy, dtype=np.float_)
    xy = np.swapaxes(xy, 0, 1)

    # get the 3D boundary
    z = np.full(len(xy), z_min, dtype=np.float_)
    z = np.expand_dims(z, axis=1)
    xyz = np.hstack((xy, z))

    # remove the duplicated points
    xyz = xyz[:-1]

    # get the face indices
    faces = np.arange(len(xyz)+1)
    faces = np.roll(faces, 1)

    # create the polygon
    polygon = pv.PolyData(xyz, faces=faces)

    return polygon


def _get_shape_mesh(z_min, z_max, obj):
    """
    Extrude a Shapely polygon into a PyVista mesh.
    """

    # ensure that the polygon has a positive orientation
    obj = sha.geometry.polygon.orient(obj)

    # get the exterior boundary and the holes
    bnd = obj.exterior
    holes = obj.interiors

    # polygon for the external boundaries
    polygon = _get_boundary_polygon(bnd, z_min)

    # polygon for the holes
    for bnd in holes:
        polygon += _get_boundary_polygon(bnd, z_min)

    # triangulate the resulting polygon
    polygon = polygon.delaunay_2d(edge_source=polygon)

    # extrude the polygon into a 3D mesh
    mesh = polygon.extrude((0, 0, z_max-z_min), capping=True)

    return mesh


def _get_shape_single(shape_type, shape_data):
    """
    Get a Shapely object for different shapes:
        - a trace (multi-segment line)
        - a pad (filled disc)
        - a polygon
    """

    # get the shape
    if shape_type == "pad":
        buffer = 0.5*shape_data["diameter"]
        obj = sha.geometry.MultiPoint(shape_data["coord"])
    elif shape_type == "trace":
        buffer = 0.5*shape_data["width"]
        obj = sha.geometry.LineString(shape_data["coord"])
    elif shape_type == "polygon":
        buffer = shape_data["buffer"]
        obj = sha.geometry.Polygon(shape_data["coord_shell"], holes=shape_data["coord_holes"])
    else:
        raise ValueError("invalid shape type")

    # add a buffer with a given thickness around the shape
    if buffer is not None:
        obj = obj.buffer(buffer, cap_style="round", join_style="round")

    # check if valid
    if obj.is_empty:
        raise RuntimeError("invalid shape: shape is empty")
    if not obj.is_valid:
        raise RuntimeError("invalid shape: geometry is ill-formed")

    return obj


def _get_voxelize_shape(n, xyz_min, xyz_max, obj):
    """
    Voxelize a shape with given bounds and voxel numbers.
    """

    # get the bounds for the voxelization
    (nx, ny, nz) = n
    (x_min, y_min, z_min) = xyz_min
    (x_max, y_max, z_max) = xyz_max

    # coordinate transformation for the bounds
    transform = rat.from_bounds(x_min, y_max, x_max, y_min, nx, ny)

    # rasterize the data
    idx_shape = raf.rasterize([obj], out_shape=(ny, nx), transform=transform)

    # get the boolean matrix with the voxelization result
    idx_shape = np.swapaxes(idx_shape, 0, 1)

    # find the 2D indices
    idx_shape = idx_shape.flatten(order="F")
    idx_shape = np.flatnonzero(idx_shape)

    return idx_shape


def _get_idx_voxel(n, idx_shape, stack_idx):
    """
    Find the 3D voxel indices from the 2D image indices.
    The number of layers to be added is arbitrary.
    """

    # get the shape size
    (nx, ny, nz) = n

    # init voxel indices
    idx_voxel = np.empty(0, dtype=np.int_)

    # convert image indices into voxel indices
    for idx in stack_idx:
        # convert indices
        idx_tmp = idx*nx*ny+idx_shape

        # add the indices to the array
        idx_voxel = np.append(idx_voxel, idx_tmp)

    return idx_voxel


def _get_shape_assemble(geometry_shape, tag, tol):
    """
    Assemble the shapes (for a specified layer).
    """

    # init the shape list
    obj_add = []
    obj_sub = []

    # find the shapes
    for geometry_shape_tmp_tmp in geometry_shape:
        # extract the data
        shape_layer = geometry_shape_tmp_tmp["shape_layer"]
        shape_operation = geometry_shape_tmp_tmp["shape_operation"]
        shape_type = geometry_shape_tmp_tmp["shape_type"]
        shape_data = geometry_shape_tmp_tmp["shape_data"]

        # add the shape
        if tag in shape_layer:
            # get the shape
            obj = _get_shape_single(shape_type, shape_data)

            # check the shape
            if not obj.is_valid:
                raise RuntimeError("invalid shape: geometry is ill-formed")

            # add to the list
            if shape_operation == "add":
                obj_add.append(obj)
            elif shape_operation == "sub":
                obj_sub.append(obj)
            else:
                raise ValueError("invalid shape type")

    # assemble the shapes
    obj_add = sha.unary_union(obj_add)
    obj_sub = sha.unary_union(obj_sub)
    obj = sha.difference(obj_add, obj_sub)

    # simplify the shape
    if tol is not None:
        obj = sha.simplify(obj, tol, preserve_topology=False)

    return obj


def _get_shape_layer(geometry_shape, stack_tag, tol):
    """
    Assemble the shapes (for all the layers).
    """

    # init list
    layer_list = []
    obj_list = []

    # get the shapes
    for tag in stack_tag:
        # get the assembled shape
        obj = _get_shape_assemble(geometry_shape, tag, tol)

        # check that the shape is valid
        if not obj.is_valid:
            raise RuntimeError("invalid shape: geometry is ill-formed")

        # add the object
        if not obj.is_empty:
            layer_list.append(tag)
            obj_list.append(obj)

    return layer_list, obj_list


def _get_shape_obj(geometry_shape, stack_tag, tol):
    """
    Create the shapes and set the position in the layer stack.
    Find the bounding box for all the shapes (minimum and maximum coordinates).
    """

    # init shape list
    shape_obj = []

    # init the coordinate (minimum and maximum coordinates)
    xy_min = np.full(2, +np.inf, dtype=np.float_)
    xy_max = np.full(2, -np.inf, dtype=np.float_)

    # create the shapes and find the bounding box
    for tag, geometry_shape_tmp in geometry_shape.items():
        # get the shape (divided per layer)
        (layer_list, obj_list) = _get_shape_layer(geometry_shape_tmp, stack_tag, tol)

        # find the layer position and add the objects
        for layer, obj in zip(layer_list, obj_list):
            # get the stack position
            idx = stack_tag.index(layer)

            # find the bounds
            (x_min, y_min, x_max, y_max) = obj.bounds
            tmp_min = np.array((x_min, y_min), dtype=np.float_)
            tmp_max = np.array((x_max, y_max), dtype=np.float_)

            # update the bounds
            xy_min = np.minimum(xy_min, tmp_min)
            xy_max = np.maximum(xy_max, tmp_max)

            # assign the object
            shape_obj.append({"tag": tag, "idx": idx, "obj": obj})

    return shape_obj, xy_min, xy_max


def _get_layer_stack(layer_stack, dz, cz):
    """
    Parse the layer stack (defined with a list).
    Return the number of layers and the coordinates of the layers.
    """

    # init the list
    stack_n = []
    stack_idx = []
    stack_tag = []

    # get a list with all the layers and the corresponding names
    for layer_stack_tmp in layer_stack:
        # extract the data
        n_layer = layer_stack_tmp["n_layer"]
        tag_layer = layer_stack_tmp["tag_layer"]

        # find the layer indices
        idx_layer = np.arange(np.sum(stack_n), np.sum(stack_n)+n_layer, dtype=np.int_)

        # append the results
        stack_n.append(n_layer)
        stack_idx.append(idx_layer)
        stack_tag.append(tag_layer)

    # get the positions
    stack_pos = np.append(0, np.cumsum(stack_n))
    stack_pos = stack_pos-np.sum(stack_n)/2

    # scale the dimension
    stack_pos = stack_pos*dz+cz

    return stack_pos, stack_idx, stack_tag


def _get_domain_def(n, d, c, geometry_shape, stack_idx, shape_obj):
    """
    Voxelize the shapes and assign the indices to a dict.
    """

    # get the voxelization bounds
    xyz_min = c-(n*d)/2
    xyz_max = c+(n*d)/2

    # init the domain dict
    domain_def = {}
    for tag in geometry_shape:
        domain_def[tag] = np.empty(0, np.int_)

    # voxelize the shapes
    for shape_obj_tmp in shape_obj:
        # extract the data
        tag = shape_obj_tmp["tag"]
        obj = shape_obj_tmp["obj"]
        idx = shape_obj_tmp["idx"]

        # voxelize and get the indices
        idx_shape = _get_voxelize_shape(n, xyz_min, xyz_max, obj)
        idx_voxel = _get_idx_voxel(n, idx_shape, stack_idx[idx])

        # append the indices into the corresponding domain
        domain_def[tag] = np.append(domain_def[tag], idx_voxel)

    # remove duplicates
    for tag, idx_voxel in domain_def.items():
        idx_voxel = np.unique(idx_voxel)
        LOGGER.debug("%s: n_voxel = %d" % (tag, len(idx_voxel)))
        domain_def[tag] = idx_voxel

    return domain_def


def _get_voxel_size(dx, dy, dz, stack_pos, xy_max, xy_min):
    """
    Get the parameters (size, dimension, and center) of the voxel structure.
    """

    # get the voxel geometry
    (x_min, y_min) = xy_min
    (x_max, y_max) = xy_max

    # get the z dimension
    z_min = np.min(stack_pos)
    z_max = np.max(stack_pos)

    # get the arrays
    d = np.array([dx, dy, dz], dtype=np.float_)
    xyz_min = np.array([x_min, y_min, z_min], dtype=np.float_)
    xyz_max = np.array([x_max, y_max, z_max], dtype=np.float_)

    # geometry size
    c = (xyz_max+xyz_min)/2

    # extract the number of voxels and the voxel size
    n = np.rint((xyz_max-xyz_min)/d)
    d = (xyz_max-xyz_min)/n

    # cast data
    d = d.astype(np.float_)
    n = n.astype(np.int_)

    # check voxel validity
    if not np.all(d > 0):
        RuntimeError("invalid voxel dimension: should be positive")
    # check voxel validity
    if not np.all(n > 0):
        RuntimeError("invalid voxel number: should be positive")

    return n, d, c


def _get_merge_shape(stack_pos, shape_obj):
    """
    Transform all the 2D vector shapes into 3D meshes.
    Merge all the meshes into a single mesh.
    The resulting mesh represent the original geometry.
    This mesh can be used to assess the quality of the voxelization.
    """

    # list for storing the meshes
    mesh_list = []

    # merge all the shapes
    for shape_obj_tmp in shape_obj:
        # extract the data
        obj = shape_obj_tmp["obj"]
        idx = shape_obj_tmp["idx"]

        # get the coordinates
        z_min = stack_pos[idx+0]
        z_max = stack_pos[idx+1]

        # transform the shapes into meshes
        if isinstance(obj, sha.Polygon):
            mesh = _get_shape_mesh(z_min, z_max, obj)
            mesh_list.append(mesh)
        elif isinstance(obj, sha.MultiPolygon):
            for obj_tmp in obj.geoms:
                mesh = _get_shape_mesh(z_min, z_max, obj_tmp)
                mesh_list.append(mesh)
        else:
            raise ValueError("invalid shape type")

    # assemble the meshes
    reference = pv.MultiBlock(mesh_list)
    reference = reference.combine().extract_surface()

    return reference


def get_mesh(param, layer_stack, geometry_shape):
    """
    Transform a series of 2D vector shapes into a 3D voxel structure.
    The 3D voxel structure is constructed from:
        - a dict containing the 2D shapes
        - a list containing the layer stack of shapes
    """

    # extract the data
    dx = param["dx"]
    dy = param["dy"]
    dz = param["dz"]
    cz = param["cz"]
    tol = param["tol"]
    xy_min = param["xy_min"]
    xy_max = param["xy_max"]

    # parse layers
    LOGGER.debug("parse the layers")
    (stack_pos, stack_idx, stack_tag) = _get_layer_stack(layer_stack, dz, cz)

    # create the shapes
    LOGGER.debug("create the shapes")
    (shape_obj, xy_min_obj, xy_max_obj) = _get_shape_obj(geometry_shape, stack_tag, tol)

    # if provided, the user specified bounds are used, otherwise the STL bounds
    if xy_min is not None:
        xy_min = np.array(xy_min, np.float_)
    else:
        xy_min = xy_min_obj
    if xy_max is not None:
        xy_max = np.array(xy_max, np.float_)
    else:
        xy_max = xy_max_obj

    # geometry size
    LOGGER.debug("get the voxel size")
    (n, d, c) = _get_voxel_size(dx, dy, dz, stack_pos, xy_max, xy_min)

    # check total size
    nv = np.prod(n)
    if (VOXEL_TOTAL is not None) and (nv > VOXEL_TOTAL):
        raise RuntimeError("invalid size of the voxel structure: %d" % nv)

    # init domain definition dict
    domain_def = {}
    for tag in geometry_shape:
        domain_def[tag] = np.empty(0, np.int_)

    # voxelize the shapes and get the indices
    LOGGER.debug("voxelize the shapes")
    domain_def = _get_domain_def(n, d, c, domain_def, stack_idx, shape_obj)

    # check used size
    nv = sum(len(idx) for idx in domain_def.values())
    if (VOXEL_USED is not None) and (nv > VOXEL_USED):
        raise RuntimeError("invalid number of used voxels: %d" % nv)

    # merge the shapes representing the original geometry
    LOGGER.debug("merge the shapes")
    reference = _get_merge_shape(stack_pos, shape_obj)

    # cast to lists
    n = n.tolist()
    d = d.tolist()
    c = c.tolist()

    # cast reference mesh
    reference = {
        "faces": np.array(reference.faces, dtype=np.int_),
        "points": np.array(reference.points, dtype=np.float_),
    }

    return n, d, c, domain_def, reference



"""
Module for transforming a series of PNG images into a 3D voxel structure.
The 2D geometry are stacked in order to create a voxel structure.

The following axis definition is used:
    - x: x-axis of the images (standard cartesian coordinate, not image coordinate)
    - y: y-axis of the images (standard cartesian coordinate, not image coordinate)
    - z: stacking dimension of the 2D geometries

The image handling is done with Pillow.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import PIL.Image as pmg
from pypeec import log
from pypeec import config

# get a logger
LOGGER = log.get_logger("STL")

# get problem size limits
VOXEL_TOTAL = config.PROBLEM_MAX_SIZE.VOXEL_TOTAL
VOXEL_USED = config.PROBLEM_MAX_SIZE.VOXEL_USED


def _get_load_image(filename_list):
    """
    Load several images into a list of tensors.
    Flip the axis in order to get standard cartesian coordinate (non-standard image coordinate).
    """

    # list for the images
    img_list = []

    # load the files
    for filename in filename_list:
        # load the image
        try:
            img = pmg.open(filename, formats=["png"])
        except OSError:
            raise RuntimeError("invalid png: invalid file content: %s" % filename)

        # cast to array
        img = img.convert("RGBA")
        img = np.array(img, dtype=np.uint8)

        # transform from image coordinate to cartesian coordinate
        img = np.swapaxes(img, 0, 1)
        img = np.flip(img, axis=1)

        # store the loaded image
        img_list.append(img)

    return img_list


def _get_idx_image(size, img, color):
    """
    Get the 2D indices of an image that correspond to a specified color.
    """

    # get the image size
    (nx, ny) = size

    # check image
    if not (img.shape == (nx, ny, 4)):
        raise RuntimeError("invalid image:  size is not compatible with the voxel structure")

    # get the color vector
    color_tmp = np.array(color, dtype=np.uint8)
    color_tmp = np.expand_dims(color_tmp, axis=(0, 1))
    if not (color_tmp.shape == (1, 1, 4)):
        raise RuntimeError("invalid color: colors should be a specified with for values")

    # find the color in the image
    idx_img = np.all(img == color_tmp, axis=2)

    # find the 2D indices
    idx_img = idx_img.flatten(order="F")
    idx_img = np.flatnonzero(idx_img)

    return idx_img


def _get_idx_voxel(size, nz, n_layer, idx_img):
    """
    Find the 3D voxel indices from the 2D image indices.
    The number of layers to be added is arbitrary.
    """

    # get the image size
    (nx, ny) = size

    # init voxel indices
    idx_voxel = np.empty(0, dtype=np.int_)

    # convert image indices into voxel indices
    for iz in range(n_layer):
        # convert indices
        idx_tmp = (nz+iz)*nx*ny+idx_img

        # add the indices to the array
        idx_voxel = np.append(idx_voxel, idx_tmp)

    return idx_voxel


def _get_domain(size, nz, n_layer, color_list, img_list):
    """
    Find the voxels indices for a list a colors and a list of images.
    Convert the 2D image indices into 3D voxel indices.
    """

    # init the index array
    idx_img = np.empty(0, dtype=np.int_)

    # get image indices (2D indices)
    for color in color_list:
        for img in img_list:
            idx_img_tmp = _get_idx_image(size, img, color)
            idx_img = np.append(idx_img, idx_img_tmp)

    # remove duplicate (between the images in the list)
    idx_img = np.unique(idx_img)

    # get voxel indices (3D indices)
    idx_voxel = _get_idx_voxel(size, nz, n_layer, idx_img)

    return idx_voxel


def _get_layer(size, nz, domain_color, domain_def, n_layer, filename_list):
    """
    Find the voxel indices for a single layer.
    A single layer can be composed of several images.
    A single layer can contain several domains.
    """

    # extract the image data as a tensor
    img_list = _get_load_image(filename_list)

    # count the number of voxel for the layer
    n_voxel = 0

    # add the indices for all the domains
    for tag, color_list in domain_color.items():
        # get the indices of the voxels
        idx_voxel = _get_domain(size, nz, n_layer, color_list, img_list)

        # count the number of voxels
        n_voxel += len(idx_voxel)

        # append the indices into the corresponding domain
        domain_def[tag] = np.append(domain_def[tag], idx_voxel)

    # display the layer size
    LOGGER.debug("size = %d / n_voxels = %d" % (n_layer, n_voxel))

    # update the layer stack
    nz += n_layer

    return nz, domain_def


def get_mesh(param, domain_color, layer_stack):
    """
    Transform a series of PNG images into a 3D voxel structure.
    The 3D voxel structure is constructed from:
        - a dict mapping the pixel colors to domains
        - a list containing the layer stack of images
    """

    # extract the data
    d = param["d"]
    c = param["c"]
    size = param["size"]

    # no reference geometry, direct voxelization
    reference = None

    # get the image size
    (nx, ny) = size

    # init the layer stack
    nz = 0

    # init domain definition dict
    domain_def = {}
    for tag in domain_color:
        domain_def[tag] = np.empty(0, np.int_)

    # add layers
    for layer_stack_tmp in layer_stack:
        # extract the data
        n_layer = layer_stack_tmp["n_layer"]
        filename_list = layer_stack_tmp["filename_list"]

        # add the layer
        (nz, domain_def) = _get_layer(size, nz, domain_color, domain_def, n_layer, filename_list)

    # assemble
    n = (nx, ny, nz)

    # check total size
    nv = np.prod(n)
    if (VOXEL_TOTAL is not None) and (nv > VOXEL_TOTAL):
        raise RuntimeError("invalid size of the voxel structure: %d" % nv)

    # check used size
    nv = sum(len(idx) for idx in domain_def.values())
    if (VOXEL_USED is not None) and (nv > VOXEL_USED):
        raise RuntimeError("invalid number of used voxels: %d" % nv)

    return n, d, c, domain_def, reference



"""
Module for generating correct value for the tests.
The values are extracted from the mesher and solver results.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"


def _get_mesher(data_geom):
    """
    Get the results produced by the mesher.
    """

    # extract the data
    n_total = data_geom["voxel_status"]["n_total"]
    n_used = data_geom["voxel_status"]["n_used"]

    # assemble results
    mesher = {
        "n_total": int(n_total),
        "n_used": int(n_used),
    }

    return mesher


def _get_solver(data_sweep):
    """
    Get the results produced by the solver.
    """

    # extract the data
    freq = data_sweep["freq"]
    has_converged = data_sweep["has_converged"]
    P_tot = data_sweep["integral"]["P_tot"]
    W_tot = data_sweep["integral"]["W_tot"]

    # assemble results
    solver = {
        "freq": float(freq),
        "has_converged": bool(has_converged),
        "P_tot": float(P_tot),
        "W_tot": float(W_tot),
    }

    return solver


def generate_results(data_voxel, data_solution):
    """
    Get the results.
    """

    # extract the data
    data_geom = data_voxel["data_geom"]
    data_sweep = data_solution["data_sweep"]

    # check the mesher
    mesher = _get_mesher(data_geom)

    # check the solver
    solver = {}
    for tag, data_sweep_tmp in data_sweep.items():
        solver[tag] = _get_solver(data_sweep_tmp)

    return mesher, solver



"""
Module for estimating the condition number of sparse matrices.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import scipy.sparse.linalg as sla
from pypeec import log

# get a logger
LOGGER = log.get_logger("COND")


def _get_decomposition(mat):
    """
    Get the LU decomposition of the provided matrix.
    """

    try:
        decomposition = sla.splu(mat)
    except RuntimeError:
        decomposition = None

    return decomposition


def _get_inverse_operator(mat, decomposition):
    """
    Get an inverse operator (with LU decomposition) for the provided matrix and the Hermitian matrix.
    """

    # get the function for the linear operator (original matrix)
    def fct_matvec(rhs):
        sol = decomposition.solve(rhs, trans="N")
        return sol

    # get the function for the linear operator (transposed matrix)
    def fct_rmatvec(rhs):
        sol = decomposition.solve(rhs, trans="H")
        return sol

    # assign linear operator for inversion
    op = sla.LinearOperator(mat.shape, matvec=fct_matvec, rmatvec=fct_rmatvec, dtype=np.complex_)

    return op


def _get_condition_matrix_sub(mat, norm_options):
    """
    Compute an estimate of the condition number (norm 1) of a sparse matrix.
    """

    # get the options
    t_accuracy = norm_options["t_accuracy"]
    n_iter_max = norm_options["n_iter_max"]

    # check shape
    nnz = mat.size
    (nx, ny) = mat.shape

    # check if the matrix is empty
    if (nx, ny) == (0, 0):
        return 0.0

    # compute matrix density
    density = nnz/(nx*ny)

    # display
    LOGGER.debug("matrix size: (%d, %d)" % (nx, ny))
    LOGGER.debug("matrix elements: %d" % nnz)
    LOGGER.debug("matrix density: %.2e" % density)

    # get LU decomposition
    LOGGER.debug("compute LU decomposition")
    decomposition = _get_decomposition(mat)

    # abort if LU decomposition failed
    if decomposition is None:
        LOGGER.warning("condition estimate is infinite")
        log.set_warning(True)
        return float("inf")

    # get the inverse operator
    op = _get_inverse_operator(mat, decomposition)

    # compute the norm of the matrix inverse (estimate)
    LOGGER.debug("estimate norm of the inverse")
    nrm_inv = sla.onenormest(op, t=t_accuracy, itmax=n_iter_max)

    # compute the norm of the matrix (estimate)
    LOGGER.debug("estimate norm of the matrix")
    nrm_ori = sla.onenormest(mat, t=t_accuracy, itmax=n_iter_max)

    # compute an estimate of the condition
    LOGGER.debug("compute condition estimate")
    cond = nrm_ori*nrm_inv

    return cond


def get_condition_matrix(name, mat, norm_options):
    """
    Compute an estimate of the condition number (norm 1) of a sparse matrix.
    """

    LOGGER.debug("condition: %s" % name)
    with log.BlockIndent():
        data = _get_condition_matrix_sub(mat, norm_options)

    return data



"""
Module for extracting the impedance matrix from the solver solution.

The impedance matrix is extracted from the terminal currents and voltages.
An equation system is formed with all the terminal quantities.
The least-square solution of the system is computed.
The impedance matrix is post-processed:
    - computation of the resistance matrix
    - computation of the inductance matrix
    - computation of the quality factors
    - computation of the couplings
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import numpy.linalg as lna


def _get_idx_matrix(n_winding):
    """
    Mapping of the impedance coefficients into a vector.
    The indices (row and column) of the impedance coefficients are returned.
    The impedance matrix is symmetric and only the independent coefficients are considered.

    """

    var_idx = []
    for i in range(n_winding):
        for j in range(i+1):
            var_idx.append([i, j])

    return var_idx


def _get_assign_matrix(n_winding, var_idx, sol):
    """
    Assign the full impedance matrix from the coefficients.
    The coefficients are given in a vector.
    """

    mat = np.zeros((n_winding, n_winding), dtype=np.complex_)
    for i, (idx_i_1, idx_i_2) in enumerate(var_idx):
        mat[idx_i_1, idx_i_2] = sol[i]
        mat[idx_i_2, idx_i_1] = sol[i]

    return mat


def _get_eqn_matrix(n_winding, var_idx, I_vec):
    """
    Create the equation matrix for a given current excitation.
    """

    eqn_mat = np.zeros((n_winding, len(var_idx)), dtype=np.complex_)
    for i, (idx_i_1, idx_i_2) in enumerate(var_idx):
        eqn_mat[idx_i_1, i] = I_vec[idx_i_2]
        eqn_mat[idx_i_2, i] = I_vec[idx_i_1]

    return eqn_mat


def _get_solve_matrix(terminal):
    """
    Extract the impedance matrix of the component.

    The impedance matrix is computed with a linear equation system.
    The equation system is built with the following variable:
        - matrix: current excitations
        - right-hand side: voltage excitations
        - solution: impedance coefficients
    """

    # extract the data
    n_solution = terminal["n_solution"]
    n_winding = terminal["n_winding"]
    freq = terminal["freq"]
    V_mat = terminal["V_mat"]
    I_mat = terminal["I_mat"]

    # check size
    assert I_mat.shape == (n_winding, n_solution), "invalid solution: current matrix shape"
    assert V_mat.shape == (n_winding, n_solution), "invalid solution: voltage matrix shape"

    # get the matrix size and indices of the coefficients
    var_idx = _get_idx_matrix(n_winding)

    # init the matrices
    rhs_vec = np.zeros(0, dtype=np.complex_)
    eqn_mat = np.zeros((0, len(var_idx)), dtype=np.complex_)

    # get the matrices
    for i in range(n_solution):
        # extract the data
        V_vec = V_mat[:, i]
        I_vec = I_mat[:, i]

        # get the equation matrix
        eqn_tmp = _get_eqn_matrix(n_winding, var_idx, I_vec)

        # append the data
        rhs_vec = np.concatenate((rhs_vec, V_vec))
        eqn_mat = np.concatenate((eqn_mat, eqn_tmp))

    # check system size
    assert len(rhs_vec) >= len(var_idx), "invalid solution: number of equation is too low"

    # check system rank
    assert lna.matrix_rank(eqn_mat) == len(var_idx),  "invalid solution: system rank is insufficient"

    # extraction impedance coefficients with a least-square solution
    (sol, _, _, _) = lna.lstsq(eqn_mat, rhs_vec, rcond=None)

    # check residuum
    res_vec = np.matmul(eqn_mat, sol)-rhs_vec
    res_rms = np.sqrt(np.mean(np.abs(res_vec)**2))
    rhs_rms = np.sqrt(np.mean(np.abs(rhs_vec)**2))
    res = res_rms/rhs_rms

    # assign the coefficients to the full impedance matrix
    Z_mat = _get_assign_matrix(n_winding, var_idx, sol)

    return n_winding, n_solution, freq, res, Z_mat


def _get_coupling_matrix(n_winding, RL_mat):
    """
    Get the coupling matrix between the windings.
    """

    k_mat = np.zeros((n_winding, n_winding), dtype=np.float_)
    for i in range(n_winding):
        for j in range(n_winding):
            k_mat[i, j] = RL_mat[i, j]/np.sqrt(RL_mat[i, i]*RL_mat[j, j])
            k_mat[j, i] = RL_mat[j, i]/np.sqrt(RL_mat[i, i]*RL_mat[j, j])

    return k_mat


def _get_parse_matrix(n_winding, n_solution, freq, res, Z_mat):
    """
    Get the equivalent circuit of the component from the impedance matrix.
    """

    # angular frequency
    w = 2*np.pi*freq

    # get the resistance
    R_mat = np.real(Z_mat)

    # get the inductance
    if freq == 0:
        L_mat = np.full(Z_mat.shape, np.nan)
    else:
        L_mat = np.imag(Z_mat)/w

    # # get the quality factor
    Q_mat = (w*L_mat)/R_mat

    # get the coupling
    k_R_mat = _get_coupling_matrix(n_winding, R_mat)
    k_L_mat = _get_coupling_matrix(n_winding, L_mat)

    # assign the results
    matrix = {
        "freq": freq, "res": res,
        "n_winding": n_winding, "n_solution": n_solution,
        "Z_mat": Z_mat, "R_mat": R_mat, "L_mat": L_mat,
        "k_R_mat": k_R_mat, "k_L_mat": k_L_mat, "Q_mat": Q_mat,
    }

    return matrix


def get_extract(terminal):
    """
    Extract the equivalent circuit of a component.
    """

    # get the impedance matrix
    (n_winding, n_solution, freq, res, Z_mat) = _get_solve_matrix(terminal)

    # get the complete circuit
    matrix = _get_parse_matrix(n_winding, n_solution, freq, res, Z_mat)

    return matrix
